import glob
import os.path
import pathlib
import tempfile
import unittest

from image_label.exporters.yolo import YoloExporter
from image_label.project import Project
from image_label.common_bus import CommonBus


class TestableStatechart:
    def __init__(self, bus):
        self.bus = bus
        self.project = Project(pathlib.Path('./assets/domik.boobalp'))
        self.bus.register_item('statechart', self)


class TestExportersYolo(unittest.TestCase):
    def setUp(self):
        self.bus = CommonBus()
        self.statechart = TestableStatechart(self.bus)
        self.y = YoloExporter(self.bus)
        self.y.path = tempfile.mkdtemp(prefix='yolo_')
        self.y.options = {
            'test_percent': 10,
            'validation_percent': 10,
            'export_rect_vals': 1
        }

    def test_mk_dest_dir(self):
        expected_dir = pathlib.Path(self.y.path, 'domik').absolute()
        self.y.mk_dest_dir()
        assert os.path.exists(expected_dir)

    def test_get_datasets_len(self):
        datasets = self.y.get_datasets()
        assert len(datasets) == 3

    def test_get_datasets_lengths(self):
        expected_lengths = {
            'train': 8,
            'test': 1,
            'val': 1
        }
        assert expected_lengths == self.y.get_datasets_lengths()

    def test_extract_project_filenames(self):
        expected_filenames = [
            'img1',
            'img2',
            'img3',
            'img4',
            'img5',
            'img6',
            'img7',
            'img8',
            'img9',
            'img10'
        ]

        assert expected_filenames == self.y.extract_project_filenames()

    def test_get_filenames_for_dataset(self):
        expected_lengths = {
            'train': 8,
            'test': 1,
            'val': 1
        }

        filenames_for_dataset = self.y.get_filenames_for_dataset(expected_lengths)

        for dataset_name, expected_len in expected_lengths.items():
            assert expected_len == len(filenames_for_dataset[dataset_name])

    def test_get_datasets_each_dataset_len(self):
        datasets = self.y.get_datasets()
        assert len(datasets['train']) == 8
        assert len(datasets['test']) == 1
        assert len(datasets['val']) == 1

    def test_extract_categories(self):
        expected_categories = {
            'poly1': 0,
            'rect1': 1,
        }

        assert expected_categories == self.y.extract_categories()

    def test_convert_file_data_into_save_data(self):
        self.y.options['export_rect_vals'] = 1

        raw_data = [('img1', {"abs_path": str(pathlib.Path("./assets/domiki.png").absolute()), "figures": [{"type": "rect", "category": "rect1", "points": [[0, 0], [0.5, 0.5]]}, {"type": "poly", "category": "poly1", "points": [[0.5, 0.5], [0.5, 1], [1.0, 1.0], [1, 0.5]]}]})]
        expected_data = {'img1': {
            'abs_path': raw_data[0][1]['abs_path'],
            'data': [
                ['1', '0.25', '0.25', '0.5', '0.5']
            ]
        }}

        converted_data = self.y.convert_file_data_into_save_data(raw_data)
        assert expected_data == converted_data

        self.y.options['export_rect_vals'] = 2
        expected_data['img1']['data'].append(['0', '0.75', '0.75', '0.5', '0.5'])
        converted_data = self.y.convert_file_data_into_save_data(raw_data)
        assert expected_data == converted_data

    def test_directory_created_mk_dataset_dir_and_save_dataset_data(self):
        dataset_data = {
            'img1': {
                'abs_path': pathlib.Path('./assets/domiki.png').absolute(),
                'data': [
                    ['1', '0.25', '0.25', '0.5', '0.5'],
                    ['0', '0.75', '0.75', '0.5', '0.5']
                ]
            }
        }
        self.y.mk_dataset_dir_and_save_dataset_data('train', dataset_data)
        expected_path = pathlib.Path(self.y.path, 'domik', 'train').absolute()
        expected_categories_file_path = pathlib.Path(expected_path, 'classes.txt')
        expected_labels_dir_path = pathlib.Path(expected_path, 'labels')
        expected_labels_img1_path = pathlib.Path(expected_labels_dir_path, 'img1.txt')
        expected_images_dir_path = pathlib.Path(expected_path, 'images')
        expected_images_img1_path = pathlib.Path(expected_images_dir_path, 'img1.png')

        assert os.path.exists(expected_path)
        assert os.path.exists(expected_categories_file_path)
        assert os.path.exists(expected_labels_dir_path)
        assert os.path.exists(expected_labels_img1_path)
        assert os.path.exists(expected_images_dir_path)
        assert os.path.exists(expected_images_img1_path)

    def test_export(self):
        self.y.options['export_rect_vals'] = 2
        self.y.export()

        expected_path = pathlib.Path(self.y.path, 'domik', 'test').absolute()
        expected_categories_file_path = pathlib.Path(expected_path, 'classes.txt')
        expected_labels_dir_path = pathlib.Path(expected_path, 'labels')
        expected_images_dir_path = pathlib.Path(expected_path, 'images')

        assert os.path.exists(expected_path)
        assert os.path.exists(expected_categories_file_path)
        assert os.path.exists(expected_labels_dir_path)
        assert os.path.exists(expected_images_dir_path)

        expected_label_txt_content = '''1 0.25 0.25 0.5 0.5\n0 0.75 0.75 0.5 0.5'''
        for label_txt in glob.glob(f'{expected_labels_dir_path}/*.txt'):
            with open(label_txt, 'r') as label_txt_handle:
                assert label_txt_handle.read() == expected_label_txt_content



if __name__ == '__main__':
    unittest.main()



import copy
import math
import os.path
import pathlib
import random
import shutil
import tempfile
import tkinter
import typing

from tkinter import filedialog
from tkinter import ttk
from PIL import Image


class YoloExporter:
    def __init__(self, bus):
        self.bus = bus
        self.bus.register_item('exporters[yolo]', self)
        self.options = {}
        self.path = None
        self.raw = None
        self.categories = {}
        self.root = None

    def show_options(self):
        self.root: tkinter.Toplevel = tkinter.Toplevel()
        self.root.title('YOLO Options')
        self.root.columnconfigure(0, weight=1)
        self.root.rowconfigure(0, weight=1)

        frame = tkinter.Frame(self.root)
        frame.grid(column=0, row=0, sticky='nesw')
        frame.columnconfigure(0, weight=1)
        frame.rowconfigure(0, weight=1)
        frame.rowconfigure(1, weight=1)
        frame.rowconfigure(2, weight=1)
        frame.rowconfigure(3, weight=1)

        label1 = tkinter.LabelFrame(frame, text='% validation')
        validation_percent = tkinter.StringVar(value='0')
        entry1 = tkinter.Entry(label1, textvariable=validation_percent)
        label1.grid(column=0, row=0, sticky='nesw')
        entry1.grid(column=0, row=0, sticky='nesw')

        label2 = tkinter.LabelFrame(frame, text='% test')
        test_percent = tkinter.StringVar(value='0')
        entry2 = tkinter.Entry(label2, textvariable=test_percent)
        label2.grid(column=0, row=1, sticky='nesw')
        entry2.grid(column=0, row=0, sticky='nesw')

        frame3 = tkinter.Frame(frame)
        frame3.rowconfigure(0, weight=1)
        frame3.rowconfigure(1, weight=1)
        frame3.columnconfigure(0, weight=1)
        frame3.columnconfigure(1, weight=1)

        frame3.grid(column=0, row=2, sticky='nesw')

        export_rect_vals = tkinter.IntVar(value=0)
        rb1 = ttk.Radiobutton(frame3, variable=export_rect_vals, value=1, text='Export only rectangulars')
        rb2 = ttk.Radiobutton(frame3, variable=export_rect_vals, value=2, text='Export with polygons (covert polygons to rectangles)')

        rb1.grid(column=0, row=0, sticky='nesw')
        rb2.grid(column=0, row=1, sticky='nesw')

        def confirm_options():
            self.ask_for_path(
                validation_percent=validation_percent.get(),
                test_percent=test_percent.get(),
                export_rect_vals=export_rect_vals.get())

        ok_bnt = tkinter.Button(frame, text='Export to YOLO format', command=confirm_options)

        self.root.bind('<Return>', lambda _: confirm_options())
        self.root.bind('<KP_Enter>', lambda _: confirm_options())

        self.root.grab_set()

        ok_bnt.grid(column=0, row=3)

    def ask_for_path(self, validation_percent, test_percent, export_rect_vals):
        if export_rect_vals == 0:
            return
        self.path = filedialog.askdirectory()
        self.root.destroy()

        if self.path:
            self.options = {'validation_percent': validation_percent,
                            'test_percent': test_percent,
                            'export_rect_vals': export_rect_vals}

            self.export()

    def get_dest_dir_name(self):
        project_name, _ = os.path.splitext(os.path.basename(self.bus.statechart.project.path))
        return pathlib.Path(self.path, project_name).absolute()

    def mk_dest_dir(self):
        dest_dir_path = self.get_dest_dir_name()
        dest_dir_path.mkdir()

    def get_datasets_lengths(self):
        total_files = len(self.bus.statechart.project.get_files())
        percent = total_files / 100
        test_dataset_len = math.ceil(float(self.options['test_percent']) * percent)
        val_dataset_len = math.ceil(float(self.options['validation_percent']) * percent)

        return {
            'train': total_files - test_dataset_len - val_dataset_len,
            'test': test_dataset_len,
            'val': val_dataset_len
        }

    def extract_project_filenames(self):
        return [filename for filename, _ in self.bus.statechart.project.get_files()]

    def get_filenames_for_dataset(self, amount_of_files_per_dataset: typing.Dict):
        files = self.extract_project_filenames()
        result = {}
        files_already_selected = []

        assert sum(amount_of_files_per_dataset.values()) <= len(files)

        for dataset_name, dataset_files_amount in amount_of_files_per_dataset.items():
            selected_files = []

            while len(selected_files) != dataset_files_amount:
                random_file = random.choice(files)
                if random_file not in files_already_selected:
                    selected_files.append(random_file)
                    files_already_selected.append(random_file)

            result[dataset_name] = copy.deepcopy(list(set(selected_files)))

        return result

    def get_datasets(self):
        datasets_lengths = self.get_datasets_lengths()
        datasets_filenames = self.get_filenames_for_dataset(datasets_lengths)

        datasets = {
            'train': self.convert_file_data_into_save_data(
                self.bus.statechart.project.get_files(datasets_filenames['train'])
            ),
            'test': self.convert_file_data_into_save_data(
                self.bus.statechart.project.get_files(datasets_filenames['test'])
            ),
            'val': self.convert_file_data_into_save_data(
                self.bus.statechart.project.get_files(datasets_filenames['val'])
            ),
        }

        return datasets

    def extract_categories(self):
        unique_categories = sorted(set(self.bus.statechart.project.get_all_categories()))
        zipped = zip(unique_categories, range(len(unique_categories)))
        return dict(zipped)

    def convert_file_data_into_save_data(self, data):
        result = {}
        categories = self.extract_categories()

        for filename, filedata in data:
            result[filename] = {
                'abs_path': copy.copy(filedata['abs_path']),
                'data': []
            }

            for figure in filedata['figures']:
                category_id = str(categories[figure['category']])
                if figure['type'] == 'rect':
                    _converted_figure_data = self._convert_rect_data_to_yolo_export_data(
                        *figure['points'][0],
                        *figure['points'][1]
                    )
                    result[filename]['data'].append([
                        category_id,
                        *map(str, _converted_figure_data)
                    ])
                elif figure['type'] == 'poly':
                    if self.options['export_rect_vals'] == 2:
                        _converted_figure_data = self._convert_poly_data_to_yolo_export_data(
                            figure['points']
                        )
                        result[filename]['data'].append([
                            category_id,
                            *map(str, _converted_figure_data)
                        ])

                if 'transformations' in filedata and len(filedata['transformations']) > 0:
                    image = Image.open(filedata['abs_path'])
                    for transform in filedata['transformations']:
                        image = self._apply_transformation_to_image(image, transform)
                    result[filename]['abs_path'] = self._save_to_temp_file(image, result[filename]['abs_path'])

        return result

    @staticmethod
    def _apply_transformation_to_image(image: Image, transform: str) -> Image:
        if transform == 'rotate_cw':
            image = image.rotate(angle=-90, expand=True)
        elif transform == 'rotate_ccw':
            image = image.rotate(angle=90, expand=True)
        return image

    @staticmethod
    def _save_to_temp_file(image: Image, src_abs_path: str) -> str:
        _, ext = os.path.splitext(src_abs_path)

        assert len(ext) > 1

        temp_path = tempfile.mktemp(suffix=ext)
        with open(temp_path, mode='wb') as temp_image_handle:
            ext = ext.strip('.')
            image.save(temp_image_handle, format=ext)

        return temp_path

    @staticmethod
    def _convert_rect_data_to_yolo_export_data(x1, y1, x2, y2):
        min_x, max_x, min_y, max_y = min(x1, x2), max(x1, x2), min(y1, y2), max(y1, y2)
        w, h = max_x - min_x, max_y - min_y

        center_x = min_x + w / 2.0
        center_y = min_y + h / 2.0

        return center_x, center_y, w, h

    @staticmethod
    def _convert_poly_data_to_yolo_export_data(points):
        xs = [point[0] for point in points]
        ys = [point[1] for point in points]
        min_x, max_x, min_y, max_y = min(xs), max(xs), min(ys), max(ys)

        w, h = max_x - min_x, max_y - min_y

        center_x = min_x + w / 2.0
        center_y = min_y + h / 2.0

        return center_x, center_y, w, h

    def mk_dataset_dir_and_save_dataset_data(self, dataset_name, dataset_data):
        dataset_dir = pathlib.Path(self.get_dest_dir_name(), dataset_name)
        dataset_dir.mkdir(parents=True)

        categories = self.extract_categories()
        categories_file_path = pathlib.Path(dataset_dir, 'classes.txt')
        with open(categories_file_path.absolute(), 'w') as categories_handle:
            categories_handle.write('\n'.join(categories.keys()))

        labels_dir = pathlib.Path(dataset_dir, 'labels')
        labels_dir.mkdir()

        images_dir = pathlib.Path(dataset_dir, 'images')
        images_dir.mkdir()

        for filename, filedata in dataset_data.items():
            label_txt_path = pathlib.Path(labels_dir, f'{filename}.txt').absolute()
            label_txt_str = '\n'.join([' '.join(line) for line in filedata['data']])

            with open(label_txt_path, 'w') as label_txt_handle:
                label_txt_handle.write(label_txt_str)

            _, ext = os.path.splitext(filedata['abs_path'])
            image_path = pathlib.Path(images_dir, f'{filename}{ext}').absolute()
            shutil.copyfile(filedata['abs_path'], image_path.absolute())

    def export(self):
        datasets = self.get_datasets()
        self.mk_dest_dir()

        for dataset_name, dataset_data in datasets.items():
            self.mk_dataset_dir_and_save_dataset_data(dataset_name, dataset_data)


import copy
import tkinter
import typing
from tkinter import ttk

from PIL import ImageTk, Image

from image_label import helpers
from image_label.common_bus import CommonBus
from image_label.figure import Figure

class Gui:
    def __init__(self, bus: CommonBus):
        self.bus = bus
        self.bus.register_item('gui', self)

        self.image_to_load_on_canvas: Image = None
        self.image_on_canvas: ImageTk = None
        self.mouse_position_marker = None
        self.drawing_rect_point_1 = None
        self.drawing_rect_figure = None

        self.drawing_poly_points = None
        self.drawing_poly_figures = None

        self.moving_figure_point = None

        self.root = tkinter.Tk()
        self.root.title('Simple Image Label 👀')
        self.root.attributes('-zoomed', True)
        self.root.geometry(f'{self.root.winfo_screenwidth()}x{self.root.winfo_screenheight()}')

        self.main_menu = None
        self.project_menu = None
        self.figure_menu = None
        self.files_frame_treeview_menu = None
        self.figures_frame_treeview_menu = None
        self.exporters_menu = None
        self.help_menu = None

        self.root.columnconfigure(0, weight=90)
        self.root.columnconfigure(1, weight=10)
        self.root.rowconfigure(0, weight=80)
        self.root.rowconfigure(1, weight=20)

        self.drawing_frame = tkinter.Frame(self.root)
        self.files_frame = tkinter.Frame(self.root)
        self.figures_frame = tkinter.Frame(self.root)

        self.drawing_frame.grid(column=0, row=0, sticky='nesw', rowspan=2)
        self.files_frame.grid(column=1, row=0, sticky='nesw')
        self.figures_frame.grid(column=1, row=1, sticky='nesw')

        # DRAWING
        self.drawing_frame.columnconfigure(0, weight=1)
        self.drawing_frame.rowconfigure(1, weight=1)

        self.drawing_frame_commands = tkinter.Frame(self.drawing_frame)
        self.drawing_frame_commands_draw_figure_commands = tkinter.Frame(self.drawing_frame_commands)
        self.draw_rectangle_btn = tkinter.Button(self.drawing_frame_commands_draw_figure_commands, text='Draw Rectangle')
        self.draw_polygon_btn = tkinter.Button(self.drawing_frame_commands_draw_figure_commands, text='Draw Polygon')

        self.drawing_frame_commands_point_manipulation_commands = tkinter.Frame(self.drawing_frame_commands, padx=10)
        self.add_point_btn = tkinter.Button(self.drawing_frame_commands_point_manipulation_commands, text='Add Point')
        self.remove_point_btn = tkinter.Button(self.drawing_frame_commands_point_manipulation_commands, text='Remove Point')
        self.move_point_btn = tkinter.Button(self.drawing_frame_commands_point_manipulation_commands, text='Move Point')

        self.drawing_frame_commands_image_rotation_commands = tkinter.Frame(self.drawing_frame_commands)
        self.rotate_ccw_btn = tkinter.Button(self.drawing_frame_commands_image_rotation_commands, text='Image Rotate Left')
        self.rotate_cw_btn = tkinter.Button(self.drawing_frame_commands_image_rotation_commands, text='Image Rotate Right')

        self.drawing_frame_canvas = tkinter.Canvas(self.drawing_frame)

        self.drawing_frame_commands.grid(column=0, row=0, sticky='nesw')
        self.drawing_frame_commands_draw_figure_commands.grid(column=0, row=0, sticky='nesw')
        self.drawing_frame_commands_point_manipulation_commands.grid(column=1, row=0, sticky='nesw')
        self.drawing_frame_commands_image_rotation_commands.grid(column=2, row=0, sticky='nesw')

        self.draw_rectangle_btn.grid(column=0, row=0, sticky='w')
        self.draw_polygon_btn.grid(column=1, row=0, sticky='w')

        self.add_point_btn.grid(column=0, row=0, sticky='w')
        self.remove_point_btn.grid(column=1, row=0, sticky='w')
        self.move_point_btn.grid(column=2, row=0, sticky='w')
        self.rotate_ccw_btn.grid(column=0, row=0, sticky='w')
        self.rotate_cw_btn.grid(column=1, row=0, sticky='w')

        self.drawing_frame_canvas.grid(column=0, row=1, sticky='nesw')
        # END DRAWING

        # FILES
        self.files_frame.columnconfigure(0, weight=1)
        self.files_frame.columnconfigure(1, weight=1)
        self.files_frame.rowconfigure(1, weight=1)

        self.add_file_btn = tkinter.Button(self.files_frame, text='Add File')
        self.remove_file_btn = tkinter.Button(self.files_frame, text='Remove File')
        self.files_frame_treeview = ttk.Treeview(self.files_frame,
                                                 columns=['filename'],
                                                 selectmode='browse',
                                                 show='headings')
        self.files_frame_treeview.heading('filename', text='Filename')

        self.add_file_btn.grid(column=0, row=0)
        self.remove_file_btn.grid(column=1, row=0)
        self.files_frame_treeview.grid(column=0, row=1, sticky='nesw', columnspan=2)
        self.files_frame_treeview_scrollbar = ttk.Scrollbar(self.files_frame, orient='vertical')
        self.files_frame_treeview_scrollbar.grid(column=2, row=1, sticky='nesw')
        self.files_frame_treeview.config(yscrollcommand=self.files_frame_treeview_scrollbar.set)
        self.files_frame_treeview_scrollbar.config(command=self.files_frame_treeview.yview)
        # END FILES

        self.figures_frame_treeview = ttk.Treeview(self.figures_frame,
                                                   columns=['figure', 'id'],
                                                   selectmode='browse',
                                                   show='headings',
                                                   displaycolumns=('figure',))
        self.figures_frame_treeview.heading('figure', text='Figure')

        self.figures_frame.columnconfigure(0, weight=1)
        self.figures_frame.rowconfigure(0, weight=1)
        self.figures_frame_treeview.grid(column=0, row=0, sticky='nesw')
        self.figures_frame_treeview_scrollbar = ttk.Scrollbar(self.figures_frame, orient='vertical')
        self.figures_frame_treeview_scrollbar.grid(column=1, row=0, sticky='nesw')
        self.figures_frame_treeview.config(yscrollcommand=self.figures_frame_treeview_scrollbar.set)
        self.figures_frame_treeview_scrollbar.config(command=self.figures_frame_treeview.yview)

        self.init_bindings()

    def init_bindings(self):
        self.add_file_btn.configure(command=lambda: helpers.add_file_event(self.bus.statechart))
        self.remove_file_btn.configure(
            command=lambda: helpers.remove_file_event(self.bus.statechart, self.files_frame_treeview.selection()[0]))

        self.draw_rectangle_btn.configure(command=lambda: helpers.draw_rect_event(self.bus.statechart))
        self.draw_polygon_btn.configure(command=lambda: helpers.draw_poly_event(self.bus.statechart))
        self.add_point_btn.configure(command=lambda: helpers.add_point_event(self.bus.statechart))
        self.remove_point_btn.configure(command=lambda: helpers.remove_point_event(self.bus.statechart))
        self.move_point_btn.configure(command=lambda: helpers.move_point_event(self.bus.statechart))
        self.rotate_ccw_btn.configure(command=lambda: helpers.rotate_ccw_event(self.bus.statechart))
        self.rotate_cw_btn.configure(command=lambda: helpers.rotate_cw_event(self.bus.statechart))

        self.root.bind('<Control-n>', lambda _: helpers.new_project_event(self.bus.statechart))
        self.root.bind('<Control-o>', lambda _: helpers.load_project_event(self.bus.statechart))
        self.root.bind('<Control-s>', lambda _: helpers.save_project_event(self.bus.statechart))

        self.root.bind('<Control-i>', lambda _: helpers.add_file_event(self.bus.statechart))

        self.root.bind('<Control-r>', lambda _: helpers.draw_rect_event(self.bus.statechart))
        self.root.bind('<KeyPress-1>', lambda _: helpers.draw_rect_event(self.bus.statechart))
        self.root.bind('<Control-p>', lambda _: helpers.draw_poly_event(self.bus.statechart))
        self.root.bind('<KeyPress-2>', lambda _: helpers.draw_poly_event(self.bus.statechart))

        self.root.bind('<KeyPress-a>', lambda _: helpers.add_point_event(self.bus.statechart))
        self.root.bind('<KeyPress-x>', lambda _: helpers.remove_point_event(self.bus.statechart))
        self.root.bind('<KeyPress-g>', lambda _: helpers.move_point_event(self.bus.statechart))

        self.root.bind('<Control-z>', lambda _: helpers.undo_event(self.bus.statechart))

        self.root.bind('<F1>', lambda _: self.show_help())

        self.main_menu = tkinter.Menu(borderwidth=0)
        self.project_menu = tkinter.Menu(tearoff=False)
        self.figure_menu = tkinter.Menu(tearoff=False)
        self.exporters_menu = tkinter.Menu(tearoff=False)
        self.help_menu = tkinter.Menu(tearoff=False)

        if self.bus.exporters:
            for k, v in self.bus.exporters.items():
                self.exporters_menu.add_command(label=k, command=lambda: v.show_options())

        self.project_menu.add_command(label='New Project', accelerator='<Control-n>',
                                      command=lambda: helpers.new_project_event(self.bus.statechart))
        self.project_menu.add_command(label='Load Project', accelerator='<Control-o>',
                                      command=lambda: helpers.load_project_event(self.bus.statechart))
        self.project_menu.add_command(label='Save Project', accelerator='<Control-s>',
                                      command=lambda: helpers.save_project_event(self.bus.statechart),
                                      state='disabled')
        self.project_menu.add_separator()
        self.project_menu.add_command(label='Add File', accelerator='<Control-i>',
                                      command=lambda: helpers.add_file_event(self.bus.statechart),
                                      state='disabled')
        self.project_menu.add_command(label='Remove File', accelerator='<Delete>',
                                      command=lambda: helpers.remove_file_event(self.bus.statechart,
                                                                                self.files_frame_treeview.selection()[
                                                                                    0]),
                                      state='disabled')
        self.project_menu.add_separator()

        self.project_menu.add_command(label='Quick Categories',
                                      command=lambda: self.show_quick_categories_settings(),
                                      state='normal')

        self.project_menu.add_separator()
        self.project_menu.add_command(label='Quit',
                                      command=lambda: helpers.quit_event(self.bus.statechart))

        self.figure_menu.add_command(label='Rotate CW (Right)',
                                     command=lambda: helpers.rotate_cw_event(self.bus.statechart),
                                     accelerator='<e>',
                                     state='disabled')
        self.figure_menu.add_command(label='Rotate CCW (Left)',
                                     command=lambda: helpers.rotate_ccw_event(self.bus.statechart),
                                     accelerator='<q>',
                                     state='disabled')
        self.figure_menu.add_separator()

        self.figure_menu.add_command(label='Draw Rectangle', accelerator='<Control-r>',
                                     command=lambda: helpers.draw_rect_event(self.bus.statechart),
                                     state='disabled')
        self.figure_menu.add_command(label='Draw Polygon', accelerator='<Control-p>',
                                     command=lambda: helpers.draw_poly_event(self.bus.statechart),
                                     state='disabled')
        self.figure_menu.add_separator()

        self.figure_menu.add_command(label='Add Point',
                                     command=lambda: helpers.add_point_event(self.bus.statechart),
                                     state='disabled')
        self.figure_menu.add_command(label='Remove Point',
                                     command=lambda: helpers.remove_point_event(self.bus.statechart),
                                     state='disabled')
        self.figure_menu.add_command(label='Move Point',
                                     command=lambda: helpers.move_point_event(self.bus.statechart),
                                     state='disabled')
        self.figure_menu.add_separator()

        self.figure_menu.add_command(label='Undo', accelerator='<Control-z>',
                                     command=lambda: helpers.undo_event(self.bus.statechart),
                                     state='disabled')

        self.help_menu.add_command(label='Help', command=lambda: self.show_help())

        self.main_menu.add_cascade(label='Project', menu=self.project_menu)
        self.main_menu.add_cascade(label='Figure', menu=self.figure_menu)
        self.main_menu.add_cascade(label='Export', menu=self.exporters_menu)
        self.main_menu.add_cascade(label='Help', menu=self.help_menu)

        self.files_frame_treeview_menu = tkinter.Menu(tearoff=False)
        self.files_frame_treeview_menu.add_command(label='Add File',
                                                   command=lambda: helpers.add_file_event(self.bus.statechart))
        self.files_frame_treeview_menu.add_command(label='Remove File',
                                                   command=lambda: helpers.remove_file_event(self.bus.statechart,
                                                                                             self.files_frame_treeview.selection()[
                                                                                                 0]))

        self.figures_frame_treeview_menu = tkinter.Menu(tearoff=False)
        self.figures_frame_treeview_menu.add_command(label='Change Category',
                                                     command=lambda: self.show_popup_figure_category_rename())
        self.figures_frame_treeview_menu.add_command(label='Change Color', state='disabled')
        self.figures_frame_treeview_menu.add_separator()
        self.figures_frame_treeview_menu.add_command(label='Delete', command=lambda: self.send_figure_delete_event())

        self.files_frame_treeview.bind('<Button-1>', lambda _: self.files_frame_treeview_menu.unpost())
        self.files_frame_treeview.bind('<Button-3>', lambda _: self.show_files_frame_treeview_menu())

        self.figures_frame_treeview.bind('<Button-1>', lambda _: self.figures_frame_treeview_menu.unpost())
        self.figures_frame_treeview.bind('<Button-3>', lambda _: self.show_figures_frame_treeview_menu())

        self.root.bind('<Left>', lambda _: self.select_prev_image())
        self.root.bind('<Up>', lambda _: self.select_prev_image())
        self.root.bind('<a>', lambda _: self.select_prev_image())
        self.root.bind('<d>', lambda _: self.select_next_image())
        self.root.bind('<Right>', lambda _: self.select_next_image())
        self.root.bind('<Down>', lambda _: self.select_next_image())

        self.root.bind('<q>', lambda _: helpers.rotate_ccw_event(self.bus.statechart))
        self.root.bind('<e>', lambda _: helpers.rotate_ccw_event(self.bus.statechart))

        self.root.config(menu=self.main_menu)

    def run(self) -> None:
        self.root.mainloop()

    def update_title(self, new_name: str) -> None:
        self.root.title(f'{new_name} | Simple Image Label 👀' if new_name else 'Simple Image Label 👀')

    def set_default_pointer(self) -> None:
        self.root.config(cursor='')

    def set_drawing_pointer(self) -> None:
        self.root.config(cursor='pencil')

    def set_grab_pointer(self) -> None:
        self.root.config(cursor='fleur')

    def set_remove_pointer(self) -> None:
        self.root.config(cursor='X_cursor')

    def set_add_pointer(self) -> None:
        self.root.config(cursor='plus')

    def add_file(self, id_, filedata) -> None:
        self.files_frame_treeview.insert('', 'end', id=id_, values=(filedata['abs_path']), tags=('#files',))

    def clear_files(self) -> None:
        if self.files_frame_treeview.tag_has('#files'):
            for item in self.files_frame_treeview.get_children(''):
                self.files_frame_treeview.delete(item)

    def remove_file(self, file_id) -> None:
        if self.files_frame_treeview.tag_has('#files'):
            for item in self.files_frame_treeview.get_children():
                if item == file_id:
                    self.files_frame_treeview.delete(item)

    def clear_image(self) -> None:
        if self.figures_frame_treeview.tag_has('#image'):
            for item in self.figures_frame_treeview.get_children(''):
                self.figures_frame_treeview.delete(item)

    def clear_figures(self) -> None:
        if self.figures_frame_treeview.tag_has('#figures'):
            for item in self.figures_frame_treeview.get_children(''):
                self.figures_frame_treeview.delete(item)

    def clear_canvas(self) -> None:
        if self.drawing_frame_canvas.gettags('#draw_figures'):
            self.drawing_frame_canvas.delete('#draw_figures')

    def bind_select_image_listener(self) -> None:
        self.files_frame_treeview.bind('<Double-Button-1>', lambda _: helpers.select_image_event(self.bus.statechart, self.files_frame_treeview.selection()[0]))

    def unbind_select_image_listener(self) -> None:
        self.files_frame_treeview.unbind('<Double-Button-1>')

    def load_image_into_canvas(self, abs_path, transformations: typing.Optional[typing.List[str]] = None) -> None:
        self.clear_image()

        self.image_to_load_on_canvas = Image.open(abs_path)

        if transformations:
            for transform in transformations:
                if transform == 'rotate_cw':
                    self.image_to_load_on_canvas = self.image_to_load_on_canvas.rotate(angle=-90, expand=True)
                elif transform == 'rotate_ccw':
                    self.image_to_load_on_canvas = self.image_to_load_on_canvas.rotate(angle=90, expand=True)

        new_sizes = self._get_size_to_resize(
            i_w=self.image_to_load_on_canvas.width, i_h=self.image_to_load_on_canvas.height,
            c_w=self.drawing_frame_canvas.winfo_width(), c_h=self.drawing_frame_canvas.winfo_height()
        )
        self.image_to_load_on_canvas = self.image_to_load_on_canvas.resize(new_sizes, resample=Image.Resampling.NEAREST)
        self.image_on_canvas = ImageTk.PhotoImage(image=self.image_to_load_on_canvas)
        self.drawing_frame_canvas.create_image(
            self.drawing_frame_canvas.winfo_width() // 2,
            self.drawing_frame_canvas.winfo_height() // 2,
            image=self.image_on_canvas,
            tags=('#image',)
        )

    @staticmethod
    def _get_size_to_resize(i_w: int, i_h: int, c_w: int, c_h: int) -> typing.List[int]:
        result = [1, 1]

        if i_w / i_h >= 1:
            scale = c_w / i_w
            result = [i_w * scale, i_h * scale]
        else:
            scale = c_h / i_h
            result = [i_w * scale, i_h * scale]

        result = [helpers.clamp(0, c_w, result[0]), helpers.clamp(0, c_h, result[1])]

        return [int(pos) for pos in result]

    def enable_save_project_btn(self) -> None:
        if self.project_menu:
            self.project_menu.entryconfig('Save Project', state='normal')

    def disable_save_project_btn(self) -> None:
        if self.project_menu:
            self.project_menu.entryconfig('Save Project', state='disabled')

    def enable_add_file_btn(self) -> None:
        self.add_file_btn['state'] = 'normal'

        if self.project_menu:
            self.project_menu.entryconfig('Add File', state='normal')

    def disable_add_file_btn(self) -> None:
        self.add_file_btn['state'] = 'disabled'

        if self.project_menu:
            self.project_menu.entryconfig('Add File', state='disabled')

    def enable_remove_file_btn(self) -> None:
        self.remove_file_btn['state'] = 'normal'

        if self.project_menu:
            self.project_menu.entryconfig('Remove File', state='normal')

    def disable_remove_file_btn(self) -> None:
        self.remove_file_btn['state'] = 'disabled'

        if self.project_menu:
            self.project_menu.entryconfig('Remove File', state='disabled')

    def enable_rotate_buttons(self) -> None:
        self.rotate_cw_btn['state'] = 'normal'
        self.rotate_ccw_btn['state'] = 'normal'

        if self.figure_menu:
            self.figure_menu.entryconfig('Rotate CW (Right)', state='normal')
            self.figure_menu.entryconfig('Rotate CCW (Left)', state='normal')

    def disable_rotate_buttons(self) -> None:
        self.rotate_cw_btn['state'] = 'disabled'
        self.rotate_ccw_btn['state'] = 'disabled'

        if self.figure_menu:
            self.figure_menu.entryconfig('Rotate CW (Right)', state='disabled')
            self.figure_menu.entryconfig('Rotate CCW (Left)', state='disabled')

    def enable_draw_buttons(self) -> None:
        self.draw_rectangle_btn['state'] = 'normal'
        self.draw_polygon_btn['state'] = 'normal'

        if self.figure_menu:
            self.figure_menu.entryconfig('Draw Rectangle', state='normal')
            self.figure_menu.entryconfig('Draw Polygon', state='normal')

    def disable_draw_buttons(self) -> None:
        self.draw_rectangle_btn['state'] = 'disabled'
        self.draw_polygon_btn['state'] = 'disabled'

        if self.figure_menu:
            self.figure_menu.entryconfig('Draw Rectangle', state='disabled')
            self.figure_menu.entryconfig('Draw Polygon', state='disabled')

    def enable_point_actions_buttons(self) -> None:
        self.add_point_btn['state'] = 'normal'
        self.remove_point_btn['state'] = 'normal'
        self.move_point_btn['state'] = 'normal'

        if self.figure_menu:
            self.figure_menu.entryconfig('Add Point', state='normal')
            self.figure_menu.entryconfig('Remove Point', state='normal')
            self.figure_menu.entryconfig('Move Point', state='normal')

    def disable_point_actions_buttons(self) -> None:
        self.add_point_btn['state'] = 'disabled'
        self.remove_point_btn['state'] = 'disabled'
        self.move_point_btn['state'] = 'disabled'

        if self.figure_menu:
            self.figure_menu.entryconfig('Add Point', state='disabled')
            self.figure_menu.entryconfig('Remove Point', state='disabled')
            self.figure_menu.entryconfig('Move Point', state='disabled')

    def enable_undo_action_button(self) -> None:
        if self.figure_menu:
            self.figure_menu.entryconfig('Undo', state='normal')

    def disable_undo_action_button(self) -> None:
        if self.figure_menu:
            self.figure_menu.entryconfig('Undo', state='disabled')

    def enable_quick_categories_settings_button(self) -> None:
        if self.project_menu:
            self.project_menu.entryconfig('Quick Categories', state='normal')

    def disable_quick_categories_settings_button(self) -> None:
        if self.project_menu:
            self.project_menu.entryconfig('Quick Categories', state='disabled')

    def bind_canvas_click_event(self) -> None:
        self.drawing_frame_canvas.bind('<Button-1>', lambda _e: helpers.click_canvas_event(self.bus.statechart, self.clamp_coords_in_image_area(_e.x, _e.y)))
        self.drawing_frame_canvas.bind('<Button-3>', lambda _e: helpers.right_click_canvas_event(self.bus.statechart))
        self.root.bind('<KeyPress-Escape>', lambda _e: helpers.reset_drawing_event(self.bus.statechart))

    def unbind_canvas_click_event(self) -> None:
        self.drawing_frame_canvas.unbind('<Button-1>')
        self.drawing_frame_canvas.unbind('<Button-3>')
        self.root.unbind('<KeyPress-Escape>')

    def bind_canvas_motion_rect_drawing_stage_1(self) -> None:
        self.drawing_frame_canvas.bind('<Motion>', lambda _e: self.redraw_drawing_position_marker(*self.clamp_coords_in_image_area(_e.x, _e.y)))

    def unbind_canvas_motion_rect_drawing_stage_1(self) -> None:
        self.drawing_frame_canvas.unbind('<Motion>')

    def bind_canvas_motion_rect_drawing_stage_2(self) -> None:
        self.drawing_frame_canvas.bind('<Motion>', lambda _e: self.redraw_drawing_rect_temp_figure(*self.clamp_coords_in_image_area(_e.x, _e.y)))

    def unbind_canvas_motion_rect_drawing_stage_2(self) -> None:
        self.drawing_frame_canvas.unbind('<Motion>')

    def redraw_drawing_position_marker(self, x, y) -> None:
        self.clear_mouse_position_marker()
        self.mouse_position_marker = self.drawing_frame_canvas.create_oval(
            x - 5, y - 5,
            x + 5, y + 5,
            outline='yellow', tags=('#draw_figures',))

    def clear_mouse_position_marker(self) -> None:
        if self.mouse_position_marker:
            self.drawing_frame_canvas.delete(self.mouse_position_marker)

    def redraw_drawing_rect_temp_figure(self, x, y) -> None:
        self.clear_drawing_rect_temp_figure()

        self.redraw_drawing_position_marker(x, y)
        self.drawing_rect_figure = self.drawing_frame_canvas.create_rectangle(
            self.drawing_rect_point_1[0], self.drawing_rect_point_1[1],
            x, y,
            outline='red', width=2,
            tags=('#draw_figures',))

    def clear_drawing_rect_temp_figure(self) -> None:
        if self.drawing_rect_figure:
            self.drawing_frame_canvas.delete(self.drawing_rect_figure)

    def draw_figure(self, file_id, figure_id, figure_data, highlight: bool = False, draggable: bool = False) -> None:
        figure = Figure(file_id, figure_id, figure_data, self.image_on_canvas, self.drawing_frame_canvas)
        figure.draw(highlight=highlight, draggable=draggable)

    def insert_figure_into_figures_list(self, file_id, figure_id, figure_data) -> None:
        values = (figure_data['category'], f'{file_id};{figure_id};{figure_data["category"]}')
        tags = ('#figures',)
        self.figures_frame_treeview.insert('', 'end', values=values, tags=tags)

    def bind_canvas_motion_poly_drawing(self) -> None:
        self.drawing_frame_canvas.bind('<Motion>', lambda _e: self.redraw_drawing_poly_temp_figure(_e.x, _e.y))

    def unbind_canvas_motion_poly_drawing(self) -> None:
        self.drawing_frame_canvas.unbind('<Motion>')

    def clear_drawing_poly_temp_figure(self) -> None:
        if self.drawing_poly_figures:
            for figure in self.drawing_poly_figures:
                self.drawing_frame_canvas.delete(figure)
        self.drawing_poly_figures = []

    def redraw_drawing_poly_temp_figure(self, x, y) -> None:
        self.clear_drawing_poly_temp_figure()

        self.redraw_drawing_position_marker(x, y)

        temp_points = copy.copy(self.drawing_poly_points)
        if len(temp_points) >= 1:
            figure = self.drawing_frame_canvas.create_oval(temp_points[0][0] - 5, temp_points[0][1] - 5,
                                                           temp_points[0][0] + 5, temp_points[0][1] + 5,
                                                           outline='pink', width=5, tags=('#draw_figures',))
            self.drawing_poly_figures.append(figure)
        if len(temp_points) > 1:
            temp_points.append((x, y))
            figure = self.drawing_frame_canvas.create_polygon(temp_points, fill='green', outline='yellow', tags=('#draw_figures',))
            self.drawing_poly_figures.append(figure)

    def get_image_and_canvas_sizes(self) -> typing.Tuple[int, int, int, int]:
        return (self.image_on_canvas.width(), self.image_on_canvas.height(),
                self.drawing_frame_canvas.winfo_width(), self.drawing_frame_canvas.winfo_height())

    def clamp_coords_in_image_area(self, x, y) -> typing.Tuple[float, float]:
        i_w, i_h, c_w, c_h = self.get_image_and_canvas_sizes()
        clamped_x, clamped_y = helpers.clamp_coords_in_image_area(i_w=i_w, i_h=i_h, c_w=c_w, c_h=c_h, x=x, y=y)

        return clamped_x, clamped_y

    def from_canvas_to_image_coords(self, x, y) -> typing.Tuple[float, float]:
        i_w, i_h, c_w, c_h = self.get_image_and_canvas_sizes()
        clamped_x, clamped_y = helpers.clamp_coords_in_image_area(i_w=i_w, i_h=i_h, c_w=c_w, c_h=c_h, x=x, y=y)
        rel_x, rel_y = helpers.from_canvas_to_image_coords(i_w=i_w, i_h=i_h, c_w=c_w, c_h=c_h, x=clamped_x, y=clamped_y)

        return rel_x, rel_y

    def bind_figure_selection_event(self) -> None:
        self.figures_frame_treeview.bind('<<TreeviewSelect>>', lambda _: self.send_figure_selected_event())
        self.figures_frame_treeview.bind('<Double-Button-1>', lambda _: self.show_popup_figure_category_rename())

    def bind_figure_delete_event(self) -> None:
        self.figures_frame_treeview.bind('<KeyPress-Delete>', lambda _: self.send_figure_delete_event())

    def show_figures_frame_treeview_menu(self) -> None:
        if len(self.figures_frame_treeview.selection()) > 0:
            self.figures_frame_treeview_menu.post(
                self.root.winfo_pointerx(),
                self.root.winfo_pointery()
            )

    def show_files_frame_treeview_menu(self) -> None:
        self.files_frame_treeview_menu.entryconfig('Remove File', state='normal')
        if len(self.files_frame_treeview.selection()) > 0:
            self.files_frame_treeview_menu.post(
                self.root.winfo_pointerx(),
                self.root.winfo_pointery()
            )
        else:
            self.files_frame_treeview_menu.post(
                self.root.winfo_pointerx(),
                self.root.winfo_pointery()
            )
            self.files_frame_treeview_menu.entryconfig('Remove File', state='disabled')

    def show_popup_figure_category_rename(self) -> None:
        if len(self.figures_frame_treeview.selection()) > 0:
            data = self.figures_frame_treeview.item(self.figures_frame_treeview.selection()[0])['values'][1].split(';')
            file_id, figure_id, category_name = data[0], int(data[1]), data[2]
            helpers.ask_for_category_name(self.bus.statechart, file_id, figure_id, default_val=category_name)

    def send_figure_delete_event(self) -> None:
        if len(self.figures_frame_treeview.selection()) > 0:
            values = self.figures_frame_treeview.item(self.figures_frame_treeview.selection()[0])['values']
            _, figure_id, _ = values[1].split(';')
            helpers.delete_figure_event(self.bus.statechart, int(figure_id))

    def send_figure_selected_event(self) -> None:
        if len(self.figures_frame_treeview.selection()) > 0:
            values = self.figures_frame_treeview.item(self.figures_frame_treeview.selection()[0])['values']
            _, figure_id, _ = values[1].split(';')
            helpers.figure_selected_event(self.bus.statechart, int(figure_id))

    def bind_point_move_click(self) -> None:
        self.drawing_frame_canvas.bind('<Button-1>', lambda e: self._find_closest_draggable_point(e.x, e.y))

    def unbind_point_move_click(self) -> None:
        self.drawing_frame_canvas.unbind('<Button-1>')

    def _find_closest_draggable_point(self, x, y) -> None:
        self.moving_figure_point = None
        figures = self.drawing_frame_canvas.find_closest(x, y, halo=10)
        for figure in figures:
            tags = self.drawing_frame_canvas.gettags(figure)
            for tag in tags:
                if tag.startswith('#grabbable-data='):
                    data = tag.lstrip('#grabbable-data=').split(';')
                    self.moving_figure_point = {
                        'figure_idx': int(data[1]),
                        'point_idx': int(data[2])
                    }
                    break

    def bind_point_move_motion_event(self) -> None:
        self.drawing_frame_canvas.bind('<B1-Motion>', lambda e: helpers.update_figures_point_position_event(self.bus.statechart, (e.x, e.y), self.moving_figure_point))

    def unbind_point_move_motion_event(self) -> None:
        self.drawing_frame_canvas.unbind('<B1-Motion>')

    def bind_point_remove_click(self) -> None:
        self.drawing_frame_canvas.bind('<Button-1>', lambda e: self.send_update_figure_remove_point_event(e.x, e.y))

    def unbind_point_remove_click(self) -> None:
        self.drawing_frame_canvas.unbind('<Button-1>')

    def send_update_figure_remove_point_event(self, x, y) -> None:
        self._find_closest_draggable_point(x, y)
        if self.moving_figure_point:
            helpers.update_figure_remove_point_event(self.bus.statechart, self.moving_figure_point)

    def bind_point_add_click(self) -> None:
        self.drawing_frame_canvas.bind('<Button-1>', lambda e: self._find_closest_insertable_line(e.x, e.y))

    def unbind_point_add_click(self) -> None:
        self.drawing_frame_canvas.unbind('<Button-1>')

    def _find_closest_insertable_line(self, x, y):
        self.moving_figure_point = None
        figures = self.drawing_frame_canvas.find_closest(x, y, halo=10)
        for figure in figures:
            tags = self.drawing_frame_canvas.gettags(figure)
            for tag in tags:
                if tag.startswith('#insertable-data='):
                    data = tag.lstrip('#insertable-data=').split(';')
                    insertable_line_data = {
                        'figure_idx': int(data[1]),
                        'point_idx': int(data[2])
                    }
                    helpers.update_figure_add_point_event(self.bus.statechart, (x, y), insertable_line_data)
                    break

    def select_prev_image(self) -> None:
        file_id = self._get_file_in_direction(direction=-1)
        if file_id:
            helpers.select_image_event(self.bus.statechart, file_id)

    def select_next_image(self) -> None:
        file_id = self._get_file_in_direction(direction=1)
        if file_id:
            helpers.select_image_event(self.bus.statechart, file_id)

    def _get_file_in_direction(self, direction: int) -> str:
        next_file_index = None
        next_file = None

        all_files = self.files_frame_treeview.get_children('')
        if all_files:
            if len(self.files_frame_treeview.selection()) > 0:
                current_file = self.files_frame_treeview.selection()[0]
                current_file_index = all_files.index(current_file)
                next_file_index = current_file_index + direction

                if direction < 0:
                    if current_file_index <= 0:
                        next_file_index = len(all_files) - 1
                elif direction > 0:
                    if current_file_index == len(all_files) - 1:
                        next_file_index = 0

                next_file = all_files[next_file_index]

        return next_file

    def set_files_selection(self, file_id: str) -> None:
        self.files_frame_treeview.selection_set([file_id])

    def show_help(self):
        popup = tkinter.Toplevel(self.root)
        popup.title('Help')
        popup.resizable(False, False)

        help_info = '''
'<Control-n>' create new project
'<Control-o>' open project
'<Control-s>' save project
'<Control-i>' add file
'<Control-r>' or '<KeyPress-1>' draw rectangle
'<Control-p>' or '<KeyPress-2>' draw polygon
'<KeyPress-a>' add new point to existing figure (works only for polygons)
'<KeyPress-x>' remove point from existing figure (if point on rectangle then removes entire figure)
'<KeyPress-g>' move point to another location
'<Control-z>' undo last action only for current file
'<Double-Left-Click>' on file item on files list select file as active
'<Left-Click>' on figure item on figures list select figure on canvas
'<Right-Click>' on figure item on figures list show context menu
'<Left-Arrow>' or '<Up-Arrow>' select previous file
'<Right-Arrow>' or '<Down-Arrow>' select next file
'<KePress-q>' rotate CCW (Left)
'<KePress-e>' rotate CW (Right) 
        '''

        title_text = 'Simple Image Label (https://github.com/vkalyvayut-roboty-a-ne-chelovek/image-label)'
        title = tkinter.Label(popup, text=title_text, justify='center', pady=10)
        title.grid(column=0, row=0, sticky='nesw')

        lbl = tkinter.Label(popup, text=help_info, justify='left', padx=20)
        lbl.grid(column=0, row=1, sticky='nesw')

        ok_btn = tkinter.Button(popup, text='OK', command=lambda: popup.destroy())
        ok_btn.grid(column=0, row=2, sticky='ns')

        popup.grab_set()

        popup.bind('<Escape>', lambda _: popup.destroy())

    def show_quick_categories_settings(self):
        w = tkinter.Toplevel(self.root)
        w.title('Quick Categories')
        w.resizable(False, False)

        w.rowconfigure(0, weight=1)
        w.rowconfigure(1, weight=1)
        w.columnconfigure(0, weight=1)

        container = tkinter.Frame(w)
        container.grid(column=0, row=1, sticky='nesw')
        container.rowconfigure(0, weight=1)
        container.columnconfigure(0, weight=1)

        categories = {}

        def add_new_category_action(category_name: typing.Optional[str] = None):
            cat_idxs = list(categories.keys())
            cat_idx = (cat_idxs[-1] + 1) if len(cat_idxs) > 0 else 0
            cat = tkinter.Frame(container)
            cat.grid(column=0, row=cat_idx, sticky='ew')
            cat.rowconfigure(0, weight=1)
            cat.columnconfigure(0, weight=1)

            input_var = tkinter.StringVar(value=category_name)
            input_ = tkinter.Entry(cat, textvariable=input_var)

            input_.grid(column=0, row=0, sticky='nesw')

            def delete_category():
                cat_data = categories.pop(cat_idx)
                cat_data['container'].destroy()

            del_btn = tkinter.Button(cat, text='-', command=delete_category)
            del_btn.grid(column=1, row=0)

            categories[cat_idx] = {
                'container': cat,
                'var': input_var
            }

        def update_quick_categories_action():
            new_categories = [copy.copy(cat['var'].get()) for cat in categories.values()]
            helpers.update_quick_categories(self.bus.statechart, new_categories)
            w.destroy()

        add_new_category_btn = tkinter.Button(w, text='+', command=add_new_category_action)
        add_new_category_btn.grid(column=0, row=0, sticky='nesw')

        save_btn = tkinter.Button(w, text='Update Quick Categories', command=update_quick_categories_action)
        save_btn.grid(column=0, row=2, sticky='ew')

        existing_categories = self.bus.statechart.project.get_quick_categories()
        if existing_categories:
            for cat in existing_categories:
                add_new_category_action(cat)

        w.grab_set()

    def reset_temp_drawing(self):
        self.drawing_poly_points = []

    def append_temp_poly_drawing_point(self, point):
        self.drawing_poly_points.append(point)

    def reset_temp_rect_drawing(self):
        self.drawing_rect_point_1 = None

    def set_temp_rect_drawing_point(self, point):
        self.drawing_rect_point_1 = point


class PlaceholderGui(Gui):
    def __init__(self, bus: CommonBus):
        self.bus = bus
        self.bus.register_item('gui', self)

    def run(self) -> None:
        pass

    def update_title(self, new_name: str) -> None:
        pass

    def set_default_pointer(self) -> None:
        pass

    def set_drawing_pointer(self) -> None:
        pass

    def set_grab_pointer(self) -> None:
        pass

    def set_remove_pointer(self) -> None:
        pass

    def set_add_pointer(self) -> None:
        pass

    def add_file(self, id_, filedata) -> None:
        pass

    def clear_files(self) -> None:
        pass

    def remove_file(self, file_id) -> None:
        pass

    def clear_image(self) -> None:
        pass

    def clear_figures(self) -> None:
        pass

    def clear_canvas(self) -> None:
        pass

    def bind_select_image_listener(self) -> None:
        pass

    def unbind_select_image_listener(self) -> None:
        pass

    def load_image_into_canvas(self, abs_path, transformations: typing.Optional[typing.List[str]] = None) -> None:
        pass

    def enable_save_project_btn(self) -> None:
        pass

    def disable_save_project_btn(self) -> None:
        pass

    def enable_add_file_btn(self) -> None:
        pass

    def disable_add_file_btn(self) -> None:
        pass

    def enable_remove_file_btn(self) -> None:
        pass

    def disable_remove_file_btn(self) -> None:
        pass

    def enable_rotate_buttons(self) -> None:
        pass

    def disable_rotate_buttons(self) -> None:
        pass

    def enable_draw_buttons(self) -> None:
        pass

    def disable_draw_buttons(self) -> None:
        pass

    def enable_point_actions_buttons(self) -> None:
        pass

    def disable_point_actions_buttons(self) -> None:
        pass

    def enable_undo_action_button(self) -> None:
        pass

    def disable_undo_action_button(self) -> None:
        pass

    def enable_quick_categories_settings_button(self) -> None:
        pass

    def disable_quick_categories_settings_button(self) -> None:
        pass

    def bind_canvas_click_event(self) -> None:
        pass

    def unbind_canvas_click_event(self) -> None:
        pass

    def bind_canvas_motion_rect_drawing_stage_1(self) -> None:
        pass

    def unbind_canvas_motion_rect_drawing_stage_1(self) -> None:
        pass

    def bind_canvas_motion_rect_drawing_stage_2(self) -> None:
        pass

    def unbind_canvas_motion_rect_drawing_stage_2(self) -> None:
        pass

    def redraw_drawing_position_marker(self, x, y) -> None:
        pass

    def clear_mouse_position_marker(self) -> None:
        pass

    def redraw_drawing_rect_temp_figure(self, x, y) -> None:
        pass

    def clear_drawing_rect_temp_figure(self) -> None:
        pass

    def draw_figure(self, file_id, figure_id, figure_data, highlight: bool = False, draggable: bool = False) -> None:
        pass

    def insert_figure_into_figures_list(self, file_id, figure_id, figure_data) -> None:
        pass

    def bind_canvas_motion_poly_drawing(self) -> None:
        pass

    def unbind_canvas_motion_poly_drawing(self) -> None:
        pass

    def clear_drawing_poly_temp_figure(self) -> None:
        pass

    def redraw_drawing_poly_temp_figure(self, x, y) -> None:
        pass

    def get_image_and_canvas_sizes(self) -> typing.Tuple[int, int, int, int]:
        return 0, 0, 1, 1

    def clamp_coords_in_image_area(self, x, y) -> typing.Tuple[float, float]:
        return 0.0, 0.0

    def from_canvas_to_image_coords(self, x, y) -> typing.Tuple[float, float]:
        return 0.0, 0.0

    def bind_figure_selection_event(self) -> None:
        pass

    def bind_figure_delete_event(self) -> None:
        pass

    def show_figures_frame_treeview_menu(self) -> None:
        pass

    def show_files_frame_treeview_menu(self) -> None:
        pass

    def show_popup_figure_category_rename(self) -> None:
        pass

    def send_figure_delete_event(self) -> None:
        pass

    def send_figure_selected_event(self) -> None:
        pass

    def bind_point_move_click(self) -> None:
        pass

    def unbind_point_move_click(self) -> None:
        pass

    def bind_point_move_motion_event(self) -> None:
        pass

    def unbind_point_move_motion_event(self) -> None:
        pass

    def bind_point_remove_click(self) -> None:
        pass

    def unbind_point_remove_click(self) -> None:
        pass

    def send_update_figure_remove_point_event(self, x, y) -> None:
        pass

    def bind_point_add_click(self) -> None:
        pass

    def unbind_point_add_click(self) -> None:
        pass

    def select_prev_image(self) -> None:
        pass

    def select_next_image(self) -> None:
        pass

    def set_files_selection(self, file_id: str) -> None:
        pass

    def show_help(self):
        pass

    def show_quick_categories_settings(self):
        pass



from tkinter import Canvas
from PIL import ImageTk
from image_label.helpers import from_image_to_canvas_coords


class Figure:
    def __init__(self, file_id, figure_id, figure_data, image_on_canvas: ImageTk, canvas: Canvas):
        self.file_id = file_id
        self.figure_id = figure_id
        self.figure_data = figure_data
        self.image_on_canvas = image_on_canvas
        self.canvas = canvas

        self._points = self._convert_coords()

    def _convert_coords(self):
        i_w, i_h = self.image_on_canvas.width(), self.image_on_canvas.height()
        c_w, c_h = self.canvas.winfo_width(), self.canvas.winfo_height()
        return [from_image_to_canvas_coords(i_w, i_h, c_w, c_h, *point) for point in self.figure_data['points']]

    def _make_figure_style(self, as_line=False):
        color = self.figure_data['color'] if 'color' in self.figure_data else 'red'
        if as_line:
            return {
                'fill': color,
                'width': 3
            }
        else:
            return {
                'fill': color,
                'outline': color,
                'width': 3
            }

    def draw(self, highlight=False, draggable=False):
        if self.figure_data['type'] == 'rect':
            if highlight:
                self._draw_rect_figure()
            else:
                self._draw_rect_as_polylines()
        elif self.figure_data['type'] == 'poly':
            if highlight:
                self._draw_poly_figure()
            else:
                self._draw_poly_as_polylines()

        if draggable:
            for point_id, point_data in enumerate(self._points):
                tags = ('#draw_figures', '#grabbable', f'#grabbable-data={self.file_id};{self.figure_id};{point_id}')
                self.canvas.create_oval(point_data[0] - 5, point_data[1] - 5,
                                        point_data[0] + 5, point_data[1] + 5,
                                        fill='pink', width=5,
                                        tags=tags)

    def _draw_rect_figure(self):
        tags = ('#draw_figures',)
        self.canvas.create_rectangle(self._points[0][0], self._points[0][1],
                                     self._points[1][0], self._points[1][1],
                                     tags=tags, **self._make_figure_style())

    def _draw_rect_as_polylines(self):
        tags = ('#draw_figures',)
        style = self._make_figure_style(as_line=True)
        point_1 = self._points[0][0], self._points[0][1]
        point_2 = self._points[1][0], self._points[0][1]
        point_3 = self._points[1][0], self._points[1][1]
        point_4 = self._points[0][0], self._points[1][1]

        self.canvas.create_line(point_1, point_2, tags=tags, **style)
        self.canvas.create_line(point_2, point_3, tags=tags, **style)
        self.canvas.create_line(point_3, point_4, tags=tags, **style)
        self.canvas.create_line(point_4, point_1, tags=tags, **style)

    def _draw_poly_figure(self):
        tags = ('#draw_figures',)
        self.canvas.create_polygon(self._points, tags=tags, **self._make_figure_style())

    def _draw_poly_as_polylines(self):
        style = self._make_figure_style(as_line=True)
        p_idx = 0
        while p_idx < len(self._points):
            tags = ('#draw_figures', '#insertable', f'#insertable-data={self.file_id};{self.figure_id};{p_idx}')
            self.canvas.create_line(self._points[p_idx - 1], self._points[p_idx], tags=tags, **style)
            p_idx += 1



import copy
import json
import random
import sys
import typing
import tkinter
from tkinter import filedialog
from tkinter import messagebox
from tkinter import ttk

from miros import Event
from miros import signals
from miros import ActiveObject


def read_project_file_from_path(path_to_project_file: str) -> typing.Dict:
    result = None
    with open(path_to_project_file, 'r') as f:
        result = json.loads(f.read())
    return result


def ask_for_load_project_path() -> str:
    return filedialog.askopenfilename(filetypes=[('Booba Label Project', '.boobalp')])


def ask_for_save_project_path() -> str:
    return filedialog.asksaveasfilename(filetypes=[('Booba Label Project', '.boobalp')])


def ask_for_add_file_paths() -> typing.List[str]:
    return filedialog.askopenfilenames(filetypes=['PNG .png', 'JPEG .jpg', 'BMP .bmp'], multiple=True)


def pick_random_color():
    colors = ['#300', '#600', '#900', '#C00', '#F00', '#003', '#303', '#603', '#903', '#C03', '#F03', '#006', '#306', '#606', '#906', '#C06', '#F06', '#009', '#309', '#609', '#909', '#C09', '#F09', '#00C', '#30C', '#60C', '#90C', '#C0C', '#F0C', '#00F', '#30F', '#60F', '#90F', '#C0F', '#F0F', '#030', '#330', '#630', '#930', '#C30', '#F30', '#033', '#333', '#633', '#933', '#C33', '#F33', '#036', '#336', '#636', '#936', '#C36', '#F36', '#039', '#339', '#639', '#939', '#C39', '#F39', '#03C', '#33C', '#63C', '#93C', '#C3C', '#F3C', '#03F', '#33F', '#63F', '#93F', '#C3F', '#F3F', '#060', '#360', '#660', '#960', '#C60', '#F60', '#063', '#363', '#663', '#963', '#C63', '#F63', '#066', '#366', '#666', '#966', '#C66', '#F66', '#069', '#369', '#669', '#969', '#C69', '#F69', '#06C', '#36C', '#66C', '#96C', '#C6C', '#F6C', '#06F', '#36F', '#66F', '#96F', '#C6F', '#F6F', '#090', '#390', '#690', '#990', '#C90', '#F90', '#093', '#393', '#693', '#993', '#C93', '#F93', '#096', '#396', '#696', '#996', '#C96', '#F96', '#099', '#399', '#699', '#999', '#C99', '#F99', '#09C', '#39C', '#69C', '#99C', '#C9C', '#F9C', '#09F', '#39F', '#69F', '#99F', '#C9F', '#F9F', '#0C0', '#3C0', '#6C0', '#9C0', '#CC0', '#FC0', '#0C3', '#3C3', '#6C3', '#9C3', '#CC3', '#FC3', '#0C6', '#3C6', '#6C6', '#9C6', '#CC6', '#FC6', '#0C9', '#3C9', '#6C9', '#9C9', '#CC9', '#FC9', '#0CC', '#3CC', '#6CC', '#9CC', '#CCC', '#FCC', '#0CF', '#3CF', '#6CF', '#9CF', '#CCF', '#FCF', '#0F0', '#3F0', '#6F0', '#9F0', '#CF0', '#FF0', '#0F3', '#3F3', '#6F3', '#9F3', '#CF3', '#FF3', '#0F6', '#3F6', '#6F6', '#9F6', '#CF6', '#FF6', '#0F9', '#3F9', '#6F9', '#9F9', '#CF9', '#FF9', '#0FC', '#3FC', '#6FC', '#9FC', '#CFC', '#FFC', '#0FF', '#3FF', '#6FF', '#9FF', '#CFF']
    return random.sample(colors, 1)


def there_are_some_changes_in_project_ask_what_to_do() -> bool:
    title = 'There are some changes made to project'
    msg = 'There are some changes made to project. They all will be lost. Do you want to do it anyway?'
    return messagebox.askyesno(title=title, message=msg)


def new_project_event(s: ActiveObject) -> None:
    are_there_any_changes_made_in_project = s.check_if_there_are_any_changes_made_in_project()
    create_new_anyway = True

    if are_there_any_changes_made_in_project:
        create_new_anyway = there_are_some_changes_in_project_ask_what_to_do()

    if create_new_anyway:
        s.post_fifo(Event(signal=signals.NEW_PROJECT))


def load_project_event(s: ActiveObject, force_load_from_path: str = None) -> None:
    if force_load_from_path:
        s.post_fifo(Event(signal=signals.LOAD_PROJECT, payload=force_load_from_path))
    else:
        are_there_any_changes_made_in_project = s.check_if_there_are_any_changes_made_in_project()
        load_anyway = True

        if are_there_any_changes_made_in_project:
            load_anyway = there_are_some_changes_in_project_ask_what_to_do()

        if load_anyway:
            file_path = ask_for_load_project_path()
            if file_path:
                s.post_fifo(Event(signal=signals.LOAD_PROJECT, payload=file_path))


def quit_event(s: ActiveObject) -> None:
    are_there_any_changes_made_in_project = s.check_if_there_are_any_changes_made_in_project()
    quit_anyway = True

    if are_there_any_changes_made_in_project:
        quit_anyway = there_are_some_changes_in_project_ask_what_to_do()

    if quit_anyway:
        sys.exit(0)


def save_project_event(s: ActiveObject) -> None:
    file_path = ask_for_save_project_path()
    if file_path:
        s.post_fifo(Event(signal=signals.SAVE_PROJECT, payload=file_path))


def select_image_event(s: ActiveObject, file_id: typing.Any) -> None:
    if file_id:
        s.post_fifo(Event(signal=signals.SELECT_IMAGE, payload=file_id))


def add_file_event(s: ActiveObject, force_add_files_from_path: typing.List[str] = None) -> None:
    if force_add_files_from_path:
        s.post_fifo(Event(signal=signals.ADD_FILE, payload=force_add_files_from_path))
    else:
        file_path = ask_for_add_file_paths()
        if len(file_path) > 0:
            s.post_fifo(Event(signal=signals.ADD_FILE, payload=file_path))


def remove_file_event(s: ActiveObject, file_id: typing.Any, force: bool = False) -> None:
    if force:
        s.post_fifo(Event(signal=signals.REMOVE_FILE, payload=file_id))
    else:
        msg = f'Are you sure to remove selected file?'
        if messagebox.askyesno(title=msg, message=msg):
            s.post_fifo(Event(signal=signals.REMOVE_FILE, payload=file_id))


def draw_rect_event(s: ActiveObject) -> None:
    s.post_fifo(Event(signal=signals.DRAW_RECT))


def draw_poly_event(s: ActiveObject) -> None:
    s.post_fifo(Event(signal=signals.DRAW_POLY))


def add_point_event(s: ActiveObject) -> None:
    s.post_fifo(Event(signal=signals.ADD_POINT))


def remove_point_event(s: ActiveObject) -> None:
    s.post_fifo(Event(signal=signals.REMOVE_POINT))


def move_point_event(s: ActiveObject) -> None:
    s.post_fifo(Event(signal=signals.MOVE_POINT))


def rotate_cw_event(s: ActiveObject) -> None:
    s.post_fifo(Event(signal=signals.ROTATE_CW))


def rotate_ccw_event(s: ActiveObject) -> None:
    s.post_fifo(Event(signal=signals.ROTATE_CCW))


def update_figures_point_position_event(s: ActiveObject, coords: typing.Tuple[int, int], figure_point_data: typing.Dict) -> None:
    if figure_point_data:
        s.post_fifo(Event(signal=signals.UPDATE_FIGURE_POINT_POSITION, payload={
            'coords': coords,
            'figure_idx': figure_point_data['figure_idx'],
            'point_idx': figure_point_data['point_idx']
        }))


def update_figure_remove_point_event(s: ActiveObject, figure_point_data: typing.Dict) -> None:
    if figure_point_data:
        s.post_fifo(Event(signal=signals.UPDATE_FIGURE_REMOVE_POINT, payload={
            'figure_idx': figure_point_data['figure_idx'],
            'point_idx': figure_point_data['point_idx']
        }))


def update_figure_add_point_event(s: ActiveObject, coords: typing.Tuple[int, int], figure_point_data: typing.Dict) -> None:
    if figure_point_data:
        s.post_fifo(Event(signal=signals.UPDATE_FIGURE_INSERT_POINT, payload={
            'coords': coords,
            'figure_idx': figure_point_data['figure_idx'],
            'point_idx': figure_point_data['point_idx']
        }))


def click_canvas_event(s: ActiveObject, coords: typing.Tuple[int, int]) -> None:
    s.post_fifo(Event(signal=signals.CLICK, payload=coords))


def right_click_canvas_event(s: ActiveObject) -> None:
    s.post_fifo(Event(signal=signals.RIGHT_CLICK))


def reset_drawing_event(s: ActiveObject) -> None:
    s.post_fifo(Event(signal=signals.RESET_DRAWING))


def figure_selected_event(s: ActiveObject, id_: int) -> None:
    s.post_fifo(Event(signal=signals.FIGURE_SELECTED, payload=id_))


def delete_figure_event(s: ActiveObject, id_: int) -> None:
    s.post_fifo(Event(signal=signals.DELETE_FIGURE, payload=id_))


def init_temp_save_event(s: ActiveObject) -> None:
    s.cancel_events(Event(signal=signals.SAVE_TEMP))
    s.post_fifo(Event(signal=signals.SAVE_TEMP), deferred=True, times=1, period=300.0)


def clamp(_min, _max, cur):
    return min(_max, max(_min, cur))


# костыль, если в окне присутствует tkinter.Label
# то происходит потеря фокуса и невозможность скрытия окна
def ask_for_category_name(c: ActiveObject, file_id: str, figure_id: int, default_val: str = '', values: typing.List[str] = ()):
    if not hasattr(c.bus.gui, 'root'):
        return

    val = tkinter.StringVar(value=default_val)

    w = tkinter.Toplevel()
    w.resizable(False, False)
    w.title('Enter category name')
    w.columnconfigure(0, weight=1)
    w.rowconfigure(0, weight=1)
    w.rowconfigure(1, weight=1)
    w.rowconfigure(2, weight=1)

    def on_ok():
        if len(val.get()) > 0:
            set_figure_category_event(c, file_id, figure_id, val.get())
            w.destroy()

    label = tkinter.Label(w, text='Enter category name')
    label.grid(column=0, row=0, sticky='nesw')
    entry = ttk.Combobox(w, textvariable=val, values=values)
    entry.grid(column=0, row=1, sticky='nesw')
    btn_ok = tkinter.Button(w, text='OK', command=on_ok)
    btn_ok.grid(column=0, row=2, sticky='nesw')

    w.geometry(f'250x100+{c.bus.gui.root.winfo_pointerx()}+{c.bus.gui.root.winfo_pointery()}')
    w.bind('<KP_Enter>', lambda _: on_ok())
    w.bind('<Return>', lambda _: on_ok())

    w.grab_set()


def ask_for_quick_category_name(c: ActiveObject, file_id: str, figure_id: int, quick_categories: typing.List[str]):
    if not hasattr(c.bus.gui, 'root'):
        return

    w = tkinter.Toplevel()
    w.resizable(False, False)
    w.title('Select category name')
    w.columnconfigure(0, weight=1)
    w.rowconfigure(0, weight=1)
    w.rowconfigure(1, weight=1)

    label = tkinter.Label(w, text='Select category name')
    label.grid(column=0, row=0)
    categories_container = tkinter.Frame(w)
    categories_container.columnconfigure(0, weight=1)
    categories_container.grid(column=0, row=1, sticky='nesw')

    def on_ok(category_name: str):
        if len(category_name) > 0:
            set_figure_category_event(c, file_id, figure_id, category_name)
            w.destroy()
    # костыли для сохранения переменной категории из цикла
    make_closure = lambda category_name: lambda e: on_ok(category_name)
    make_closure_no_arg = lambda category_name: lambda: on_ok(category_name)

    for cat_idx, category in enumerate(quick_categories):
        cat_btn = tkinter.Button(categories_container,
                                 text=f'[{cat_idx}] {category}',
                                 command=make_closure_no_arg(category))
        cat_btn.grid(column=0, row=cat_idx, sticky='ew')
        if cat_idx < 10:
            w.bind(f'<KeyPress-{cat_idx}>', make_closure(category))
            w.bind(f'<KeyPress-KP_{cat_idx}>', make_closure(category))

    w.bind('<KeyPress-Escape>', lambda _: w.destroy())

    w.grab_set()


def undo_event(s: ActiveObject):
    s.post_fifo(Event(signal=signals.UNDO_HISTORY))


def set_figure_category_event(s: ActiveObject, file_id, figure_id, category_name):
    s.post_fifo(Event(signal=signals.SET_FIGURE_CATEGORY, payload={'file_id': file_id, 'figure_id': figure_id, 'category': category_name}))


def update_quick_categories(s: ActiveObject, new_categories: typing.List[str]) -> None:
    s.post_fifo(Event(signal=signals.UPDATE_QUICK_CATEGORIES, payload=new_categories))


def clamp_coords_in_image_area(i_w, i_h, c_w, c_h, x, y) -> (int, int):
    center_x, center_y = c_w / 2.0, c_h / 2.0
    min_x = center_x - i_w / 2
    max_x = center_x + i_w / 2
    min_y = center_y - i_h / 2
    max_y = center_y + i_h / 2

    clamped_x = clamp(min_x, max_x, x)
    clamped_y = clamp(min_y, max_y, y)

    return clamped_x, clamped_y


def from_canvas_to_image_coords(i_w, i_h, c_w, c_h, x, y):
    clamped_x, clamped_y = clamp_coords_in_image_area(i_w, i_h, c_w, c_h, x, y)

    center_x, center_y = c_w / 2.0, c_h / 2.0

    max_x = center_x + i_w / 2
    max_y = center_y + i_h / 2

    rel_x = 1.0 - (max_x - clamped_x) / i_w
    rel_y = 1.0 - (max_y - clamped_y) / i_h

    return rel_x, rel_y


def from_image_to_canvas_coords(i_w, i_h, c_w, c_h, x, y):
    center_x, center_y = c_w / 2.0, c_h / 2.0

    min_x = center_x - i_w / 2
    min_y = center_y - i_h / 2

    abs_x = min_x + (x * i_w)
    abs_y = min_y + (y * i_h)

    return abs_x, abs_y



import copy
import json
import os.path
import pathlib
import typing
import uuid

from image_label.history import History


class Project:
    def __init__(self, path: typing.Union[str, pathlib.Path] = None):
        self.version = 1
        self.path = path
        self.files = {}
        self.selected_file_id = None
        self.quick_categories = None

        if path is not None:
            self._load_from_path(path)

        self.history = History(self.files)

    def _load_from_path(self, path: str) -> typing.Dict:
        result = {}
        with open(path, 'r') as handle:
            data = json.loads(handle.read())
            if isinstance(data, dict):
                assert data['version'] == 1, 'incompatible version'
                if 'files' in data:
                    self.files = data['files']
                if 'quick_categories' in data:
                    self.quick_categories = data['quick_categories']

        return result

    def add_file(self, abs_path):
        file_id = str(uuid.uuid4())
        self.files[file_id] = {
            'abs_path': abs_path,
            'figures': [],
            "transformations": []}
        self.history.set_defaults(file_id, self.files[file_id])
        return file_id

    def remove_file(self, file_id):
        del self.files[file_id]

    def get_files(self, only_keys: typing.Tuple = None):
        if len(self.files) == 0:
            return []
        if only_keys is not None:
            return [(k, self.files.get(k)) for k in only_keys]
        return [(k, v) for k, v in self.files.items()]

    def select_file(self, file_id):
        self.selected_file_id = file_id

    def get_selected_file(self):
        if self.selected_file_id and self.selected_file_id in self.files:
            return self.selected_file_id, self.files[self.selected_file_id]
        return None, None

    def delete_figure(self, figure_id):
        self.history.add_snapshot(self.selected_file_id, self.files[self.selected_file_id])
        del self.files[self.selected_file_id]['figures'][figure_id]

    def save_project(self, abs_path: typing.Union[str, pathlib.Path], temp: bool = False):
        with open(abs_path, 'w+') as handle:
            data = {
                'version': self.version,
                'files': self.files,
                'quick_categories': self.quick_categories
            }
            handle.write(json.dumps(data))

        self.history.reset_history()
        if not temp:
            self.path = abs_path

    def undo_history(self):
        snapshot = self.history.pop_history(self.selected_file_id)
        self.files[self.selected_file_id] = snapshot

    def _add_figure(self, type_: str, points: typing.List[float], color: str = None, category: str = '<NOCATEGORY>'):
        self.history.add_snapshot(self.selected_file_id, self.files[self.selected_file_id])
        self.files[self.selected_file_id]['figures'].append({
            'type': type_,
            'points': copy.deepcopy(points),
            'category': category,
            'color': color,
        })
        return len(self.files[self.selected_file_id]['figures']) - 1

    def add_rectangle(self, points: typing.List[float], color: str = None, category: str = '<NOCATEGORY>'):
        return self._add_figure(type_='rect', points=points, color=color, category=category)

    def add_polygon(self, points: typing.List[float], color: str = None, category: str = '<NOCATEGORY>'):
        return self._add_figure(type_='poly', points=points, color=color, category=category)

    def update_figure_category(self, file_id, figure_id, category):
        self.files[file_id]['figures'][figure_id]['category'] = category

    def update_figure_point_position(self, file_id, figure_id, point_id, new_coords):
        self.history.add_snapshot(file_id, self.files[file_id])
        self.files[file_id]['figures'][figure_id]['points'][point_id] = new_coords

    def update_figure_remove_point(self, file_id, figure_id, point_id):
        self.history.add_snapshot(file_id, self.files[file_id])

        if self.files[file_id]['figures'][figure_id]['type'] == 'rect':
            del self.files[file_id]['figures'][figure_id]
        else:
            if len(self.files[file_id]['figures'][figure_id]['points']) <= 3:
                del self.files[file_id]['figures'][figure_id]
            else:
                del self.files[file_id]['figures'][figure_id]['points'][point_id]

    def update_figure_insert_point(self, file_id, figure_id, point_id, coords):
        self.history.add_snapshot(file_id, self.files[file_id])
        self.files[file_id]['figures'][figure_id]['points'].insert(point_id, coords)

    def get_all_categories(self):
        all_categories = []
        for _, file_data in self.get_files():
            for figure in file_data['figures']:
                all_categories.append(figure['category'])
        result = list(set(all_categories))

        return result

    def temp_save(self):
        if self.path:
            temp_save_path = pathlib.Path(f'{self.path}.tmp').absolute()
            self.save_project(temp_save_path, temp=True)

    def get_project_name(self):
        if self.path:
            return os.path.basename(self.path)

    def rotate_cw(self, file_id):
        self.history.add_snapshot(file_id, self.files[file_id])
        if 'transformations' not in self.files[file_id]:
            self.files[file_id]['transformations'] = []
        self.files[file_id]['transformations'].append('rotate_cw')

    def rotate_ccw(self, file_id):
        self.history.add_snapshot(file_id, self.files[file_id])
        if 'transformations' not in self.files[file_id]:
            self.files[file_id]['transformations'] = []
        self.files[file_id]['transformations'].append('rotate_ccw')

    def set_quick_categories(self, categories: typing.List[str]) -> None:
        self.quick_categories = categories

    def get_quick_categories(self):
        return self.quick_categories



import copy
import typing

from miros import ActiveObject
from miros import Event
from miros import return_status
from miros import signals
from miros import spy_on

from image_label.common_bus import CommonBus
from image_label import helpers
from image_label.project import Project


class Statechart(ActiveObject):
    def __init__(self, name: str, bus: CommonBus):
        super().__init__(name=name)

        self.bus = bus
        self.bus.register_item('statechart', self)

        self.project = None

        self.points = []

    def _redraw_canvas_and_figures(self, highlight_figure_id: int = None, draggable: bool = False,
                                   update_figure_list: bool = True, update_image: bool = False):
        selected_file_id, selected_file_data = self.project.get_selected_file()
        if update_figure_list:
            self.bus.gui.clear_figures()
        self.bus.gui.clear_canvas()

        if not selected_file_data:
            return

        if update_image:
            self.bus.gui.load_image_into_canvas(selected_file_data['abs_path'], selected_file_data['transformations'])

        for figure_id, figure_data in enumerate(selected_file_data['figures']):
            highlight = bool((highlight_figure_id is not None) and (highlight_figure_id == figure_id))
            self.bus.gui.draw_figure(selected_file_id, figure_id, figure_data, highlight=highlight, draggable=draggable)
            if update_figure_list:
                self.bus.gui.insert_figure_into_figures_list(selected_file_id, figure_id, figure_data)

    def ask_category_name(self, file_id, figure_id):
        categories = self.project.get_all_categories()
        default_category = categories[0] if len(categories) > 0 else ''

        if self.project.get_quick_categories() is not None:
            helpers.ask_for_quick_category_name(self, copy.copy(file_id), figure_id,
                                                quick_categories=self.project.get_quick_categories())
        else:
            helpers.ask_for_category_name(self,
                                          copy.copy(file_id),
                                          figure_id,
                                          default_val=default_category,
                                          values=categories)

    def check_if_there_are_any_undo_actions_available_and_then_change_state_of_the_undo_button(self):
        check_result = False
        if self.project:
            selected_file_id, _ = self.project.get_selected_file()
            check_result = True if selected_file_id and self.project.history.has_history(selected_file_id) else False

        if check_result:
            self.bus.gui.enable_undo_action_button()
        else:
            self.bus.gui.disable_undo_action_button()

    def check_if_there_are_any_changes_made_in_project(self):
        result = False

        if self.project:
            for file_id, _ in self.project.get_files():
                if self.project.history.has_history(file_id) and self.project.history.get_history_len(file_id) > 0:
                    result = True
                    break

        return result

    def run(self):
        self.start_at(no_project)

    def empty_project(self):
        self.project = Project()

    def load_project(self, abs_path):
        self.project = Project(abs_path)
        self.bus.gui.update_title(self.project.get_project_name())

    def on_no_project_entry(self):
        self.bus.gui.disable_save_project_btn()
        self.bus.gui.disable_add_file_btn()
        self.bus.gui.disable_remove_file_btn()
        self.bus.gui.disable_rotate_buttons()
        self.bus.gui.disable_draw_buttons()
        self.bus.gui.disable_point_actions_buttons()
        self.check_if_there_are_any_undo_actions_available_and_then_change_state_of_the_undo_button()
        self.bus.gui.disable_quick_categories_settings_button()

    def on_in_project_entry(self):
        self.bus.gui.enable_save_project_btn()
        self.bus.gui.enable_add_file_btn()
        self.bus.gui.enable_quick_categories_settings_button()

        if self.project.get_files():
            self.bus.gui.enable_remove_file_btn()

        self.bus.gui.clear_files()
        self.bus.gui.clear_figures()
        self.bus.gui.clear_canvas()

        for file_id, filedata in self.project.get_files():
            self.bus.gui.add_file(file_id, filedata)

        selected_file_id, _ = self.project.get_selected_file()

        if self.project.get_files():
            if selected_file_id:
                helpers.select_image_event(self, selected_file_id)
            else:
                for file_id, filedata in self.project.get_files():
                    helpers.select_image_event(self, file_id)
                    break

        self.bus.gui.bind_select_image_listener()
        self.bus.gui.bind_figure_delete_event()
        self.bus.gui.bind_figure_selection_event()

        helpers.init_temp_save_event(self)

    def on_in_project_exit(self):
        self.bus.gui.unbind_select_image_listener()
        self.bus.gui.clear_files()
        self.bus.gui.clear_figures()
        self.bus.gui.clear_canvas()

    def on_in_project_add_file(self, files):
        selected_file_id, _ = self.project.get_selected_file()
        for abs_path in files:
            self.project.add_file(abs_path)

        self.bus.gui.clear_files()
        for file_id, file_data in self.project.get_files():
            self.bus.gui.add_file(file_id, file_data)

        self.bus.gui.enable_remove_file_btn()

        if selected_file_id:
            helpers.select_image_event(self, selected_file_id)
        else:
            for file_id, _ in reversed(self.project.get_files()):
                helpers.select_image_event(self, file_id)
                break

    def on_in_project_select_image(self, file_id):
        self.project.select_file(file_id)
        file_id, _ = self.project.get_selected_file()

        self._redraw_canvas_and_figures(update_image=True)

        self.bus.gui.enable_draw_buttons()

        self.bus.gui.enable_rotate_buttons()

        self.bus.gui.enable_point_actions_buttons()

        self.check_if_there_are_any_undo_actions_available_and_then_change_state_of_the_undo_button()

        self.bus.gui.set_files_selection(file_id)
        helpers.reset_drawing_event(self)

        self.bus.gui.set_default_pointer()

    def on_in_project_figure_selected(self, selected_figure_id):
        self._redraw_canvas_and_figures(highlight_figure_id=selected_figure_id, update_figure_list=False)

    def on_in_project_delete_figure(self, figure_id):
        if figure_id is not None:
            self.project.delete_figure(figure_id)

        self._redraw_canvas_and_figures()

    def on_in_project_save_project(self, abs_path):
        self.project.save_project(abs_path)
        self.bus.gui.update_title(self.project.get_project_name())

    def on_in_project_undo_history(self):
        selected_file_id, _ = self.project.get_selected_file()
        if self.project.history.has_history(selected_file_id):
            self.project.undo_history()
            self._redraw_canvas_and_figures()

        self.check_if_there_are_any_undo_actions_available_and_then_change_state_of_the_undo_button()

    def on_in_project_remove_file(self, file_id):
        selected_file_id, _ = self.project.get_selected_file()

        self.bus.gui.remove_file(file_id)
        self.project.remove_file(file_id)

        files = self.project.get_files()

        if file_id == selected_file_id:
            self.bus.gui.clear_figures()
            self.bus.gui.clear_canvas()
            if len(files) > 0:
                helpers.select_image_event(self, files[-1][0])

        self.bus.gui.disable_rotate_buttons()
        self.bus.gui.disable_draw_buttons()
        if len(files) > 0:
            self.bus.gui.enable_remove_file_btn()
        else:
            self.bus.gui.disable_remove_file_btn()

    def on_in_project_set_figure_category(self, file_id, figure_id, category):
        self.project.update_figure_category(file_id, figure_id, category)

        selected_file_id, _ = self.project.get_selected_file()

        if file_id == selected_file_id:
            self._redraw_canvas_and_figures()

    def on_in_project_update_quick_categories(self, new_categories: typing.Optional[typing.List[str]]) -> None:
        self.project.set_quick_categories(new_categories)

    def on_drawing_rect_entry(self):
        self.bus.gui.bind_canvas_click_event()
        self.bus.gui.bind_canvas_motion_rect_drawing_stage_1()

        self.bus.gui.set_drawing_pointer()

        self.points = []

    def on_drawing_rect_exit(self):
        self.bus.gui.unbind_canvas_click_event()
        self.bus.gui.unbind_canvas_motion_rect_drawing_stage_1()

        self.bus.gui.set_default_pointer()

    def on_drawing_rect_reset_drawing(self):
        self.points = []
        self._redraw_canvas_and_figures()

    def on_drawing_rect_click(self, point):
        self.points.append(point)

    def on_drawing_rect_waiting_for_2_point_entry(self):
        self.bus.gui.set_temp_rect_drawing_point(self.points[0])

        self.bus.gui.bind_canvas_click_event()
        self.bus.gui.bind_canvas_motion_rect_drawing_stage_2()

        self.bus.gui.set_drawing_pointer()

    def on_drawing_rect_waiting_for_2_point_exit(self):
        self.bus.gui.unbind_canvas_motion_rect_drawing_stage_2()
        self.bus.gui.unbind_canvas_click_event()

        self.bus.gui.set_default_pointer()

        self.bus.gui.reset_temp_rect_drawing()

        self.points = []

    def on_drawing_rect_waiting_for_2_point_click(self, point):
        self.points.append(point)

        # TODO переписать в метод
        points = [self.bus.gui.from_canvas_to_image_coords(*point) for point in self.points]
        new_figure_id = self.project.add_rectangle(points=points, color=helpers.pick_random_color())
        self.points = []

        selected_file_id, _ = self.project.get_selected_file()

        self.ask_category_name(selected_file_id, new_figure_id)

        self._redraw_canvas_and_figures()

    def on_drawing_rect_waiting_for_2_point_reset_drawing(self):
        self.points = []
        selected_file_id, _ = self.project.get_selected_file()
        self._redraw_canvas_and_figures()

    def on_drawing_poly_entry(self):
        self.bus.gui.bind_canvas_click_event()
        self.bus.gui.bind_canvas_motion_poly_drawing()
        self.points = []
        self.bus.gui.drawing_poly_points = []

        self.bus.gui.set_drawing_pointer()

    def on_drawing_poly_exit(self):
        self.bus.gui.unbind_canvas_click_event()
        self.bus.gui.unbind_canvas_motion_poly_drawing()

        self.bus.gui.set_default_pointer()

    def on_drawing_poly_reset_drawing(self):
        self.points = []

        self.bus.gui.reset_temp_drawing()
        self._redraw_canvas_and_figures()

    def on_drawing_poly_click(self, point):
        self.points.append(point)
        self.bus.gui.append_temp_poly_drawing_point(point)

    def on_drawing_poly_click_check_if_have_to_finish_drawing(self) -> bool:
        result = False

        if len(self.points) > 3:
            point_start = self.points[0]
            point_finish = self.points[-1]

            if abs(point_finish[0] - point_start[0]) <= 5 and abs(point_finish[1] - point_start[1]) <= 5:
                result = True

        return result

    def on_drawing_poly_finish_drawing(self):
        self.points.pop()
        self.save_current_polygon_as_is()

    def save_current_polygon_as_is(self):
        selected_file_id, _ = self.project.get_selected_file()

        # TODO переписать в метод
        new_figure_id = self.project.add_polygon(
            points=[self.bus.gui.from_canvas_to_image_coords(*point) for point in self.points],
            color=helpers.pick_random_color()
        )

        self.ask_category_name(selected_file_id, new_figure_id)
        self._redraw_canvas_and_figures()

    def on_moving_point_entry(self):
        self._redraw_canvas_and_figures(draggable=True, update_figure_list=False)

        self.bus.gui.bind_point_move_click()
        self.bus.gui.bind_point_move_motion_event()

        self.bus.gui.set_grab_pointer()

    def on_moving_point_exit(self):
        self._redraw_canvas_and_figures(update_figure_list=False)

        self.bus.gui.unbind_point_move_click()
        self.bus.gui.unbind_point_move_motion_event()

        self.bus.gui.set_default_pointer()

    def on_moving_point_update_figure_point_position(self, figure_id, point_id, new_coords):
        selected_file_id, _ = self.project.get_selected_file()

        # TODO переписать в метод
        new_coords = self.bus.gui.from_canvas_to_image_coords(*new_coords)
        self.project.update_figure_point_position(selected_file_id, figure_id, point_id, new_coords)

        self._redraw_canvas_and_figures(draggable=True, update_figure_list=False)

    def on_removing_point_entry(self):
        self._redraw_canvas_and_figures(draggable=True)

        self.bus.gui.bind_point_remove_click()

        self.bus.gui.set_remove_pointer()

    def on_removing_point_exit(self):
        self._redraw_canvas_and_figures()

        self.bus.gui.unbind_point_remove_click()

        self.bus.gui.set_default_pointer()

    def on_removing_point_update_figure_remove_point(self, figure_id, point_id):
        selected_file_id, _ = self.project.get_selected_file()

        self.project.update_figure_remove_point(selected_file_id, figure_id, point_id)

        self._redraw_canvas_and_figures(draggable=True)

    def on_adding_point_entry(self):
        self._redraw_canvas_and_figures(draggable=True)

        self.bus.gui.bind_point_add_click()

        self.bus.gui.set_add_pointer()

    def on_adding_point_exit(self):
        self._redraw_canvas_and_figures()

        self.bus.gui.unbind_point_add_click()

        self.bus.gui.set_default_pointer()

    def on_adding_point_update_figure_insert_point(self, figure_id, point_id, coords):
        selected_file_id, _ = self.project.get_selected_file()

        # TODO переписать в метод
        new_coords = self.bus.gui.from_canvas_to_image_coords(*coords)
        self.project.update_figure_insert_point(selected_file_id, figure_id, point_id, new_coords)

        self._redraw_canvas_and_figures(draggable=True)

    def on_in_project_save_temp(self):
        self.project.temp_save()
        helpers.init_temp_save_event(self)

    def on_in_project_rotate_cw(self):
        selected_file_id, selected_file_data = self.project.get_selected_file()

        if not selected_file_data:
            return

        self.project.rotate_cw(selected_file_id)

        self._redraw_canvas_and_figures(
            update_image=True,
            update_figure_list=False)

    def on_in_project_rotate_ccw(self):
        selected_file_id, selected_file_data = self.project.get_selected_file()

        if not selected_file_data:
            return

        self.project.rotate_ccw(selected_file_id)

        self._redraw_canvas_and_figures(
            update_image=True,
            update_figure_list=False)


@spy_on
def no_project(chart: Statechart, event: Event) -> return_status:
    status = return_status.UNHANDLED

    if event.signal == signals.ENTRY_SIGNAL:
        status = return_status.HANDLED
        chart.on_no_project_entry()
    elif event.signal == signals.NEW_PROJECT:
        status = chart.trans(in_project)
        chart.empty_project()
    elif event.signal == signals.LOAD_PROJECT:
        status = chart.trans(in_project)
        chart.load_project(event.payload)
    else:
        status = return_status.SUPER
        chart.temp.fun = chart.top

    return status


@spy_on
def in_project(chart: Statechart, event: Event) -> return_status:
    status = return_status.UNHANDLED

    if event.signal == signals.ENTRY_SIGNAL:
        status = return_status.HANDLED
        chart.on_in_project_entry()
    elif event.signal == signals.EXIT_SIGNAL:
        status = return_status.HANDLED
        chart.on_in_project_exit()
    elif event.signal == signals.UNDO_HISTORY:
        status = return_status.HANDLED
        chart.on_in_project_undo_history()
    elif event.signal == signals.ADD_FILE:
        status = return_status.HANDLED
        chart.on_in_project_add_file(event.payload)
    elif event.signal == signals.REMOVE_FILE:
        status = return_status.HANDLED
        chart.on_in_project_remove_file(event.payload)
    elif event.signal == signals.SELECT_IMAGE:
        status = return_status.HANDLED
        chart.on_in_project_select_image(event.payload)
    elif event.signal == signals.SAVE_PROJECT:
        status = return_status.HANDLED
        abs_path = event.payload
        chart.on_in_project_save_project(abs_path)
    elif event.signal == signals.DRAW_RECT:
        status = chart.trans(drawing_rect)
    elif event.signal == signals.DRAW_POLY:
        status = chart.trans(drawing_poly)
    elif event.signal == signals.ADD_POINT:
        status = chart.trans(adding_point)
    elif event.signal == signals.REMOVE_POINT:
        status = chart.trans(removing_point)
    elif event.signal == signals.MOVE_POINT:
        status = chart.trans(moving_point)
    elif event.signal == signals.FIGURE_SELECTED:
        status = return_status.HANDLED
        selected_figure_id = event.payload
        chart.on_in_project_figure_selected(selected_figure_id)
    elif event.signal == signals.DELETE_FIGURE:
        status = return_status.HANDLED
        figure_id = event.payload
        chart.on_in_project_delete_figure(figure_id)
    elif event.signal == signals.SET_FIGURE_CATEGORY:
        status = return_status.HANDLED

        file_id = event.payload['file_id']
        figure_id = event.payload['figure_id']
        category = event.payload['category']

        chart.on_in_project_set_figure_category(file_id, figure_id, category)
    elif event.signal == signals.SAVE_TEMP:
        status = return_status.HANDLED
        chart.on_in_project_save_temp()
    elif event.signal == signals.ROTATE_CW:
        status = return_status.HANDLED
        chart.on_in_project_rotate_cw()
    elif event.signal == signals.ROTATE_CCW:
        status = return_status.HANDLED
        chart.on_in_project_rotate_ccw()
    elif event.signal == signals.UPDATE_QUICK_CATEGORIES:
        status = return_status.HANDLED
        chart.on_in_project_update_quick_categories(event.payload)
    else:
        status = return_status.SUPER
        chart.temp.fun = no_project

    return status


@spy_on
def drawing_rect(chart: Statechart, event: Event) -> return_status:
    status = return_status.UNHANDLED

    if event.signal == signals.ENTRY_SIGNAL:
        status = return_status.HANDLED
        chart.on_drawing_rect_entry()
    elif event.signal == signals.EXIT_SIGNAL:
        status = return_status.HANDLED
        chart.on_drawing_rect_exit()
    elif event.signal == signals.RESET_DRAWING or event.signal == signals.RIGHT_CLICK:
        status = chart.trans(in_project)
        chart.on_drawing_rect_reset_drawing()
    elif event.signal == signals.CLICK:
        status = chart.trans(drawing_rect_waiting_for_2_point)
        chart.on_drawing_rect_click(event.payload)
    else:
        status = return_status.SUPER
        chart.temp.fun = in_project

    return status


@spy_on
def drawing_rect_waiting_for_2_point(chart: Statechart, event: Event) -> return_status:
    status = return_status.HANDLED

    if event.signal == signals.ENTRY_SIGNAL:
        status = return_status.HANDLED
        chart.on_drawing_rect_waiting_for_2_point_entry()
    elif event.signal == signals.EXIT_SIGNAL:
        status = return_status.HANDLED
        chart.on_drawing_rect_waiting_for_2_point_exit()
    if event.signal == signals.CLICK:
        status = chart.trans(in_project)
        chart.on_drawing_rect_waiting_for_2_point_click(event.payload)
    elif event.signal == signals.RESET_DRAWING or event.signal == signals.RIGHT_CLICK:
        status = chart.trans(in_project)
        chart.on_drawing_rect_waiting_for_2_point_reset_drawing()
    else:
        status = return_status.SUPER
        chart.temp.fun = in_project

    return status


@spy_on
def drawing_poly(chart: Statechart, event: Event) -> return_status:
    status = return_status.UNHANDLED

    if event.signal == signals.ENTRY_SIGNAL:
        status = return_status.HANDLED
        chart.on_drawing_poly_entry()
    elif event.signal == signals.EXIT_SIGNAL:
        status = return_status.HANDLED
        chart.on_drawing_poly_exit()
    elif event.signal == signals.RESET_DRAWING:
        status = chart.trans(in_project)
        chart.on_drawing_poly_reset_drawing()
    elif event.signal == signals.CLICK:
        status = return_status.HANDLED
        chart.on_drawing_poly_click(event.payload)
        if chart.on_drawing_poly_click_check_if_have_to_finish_drawing():
            chart.on_drawing_poly_finish_drawing()
            status = chart.trans(in_project)
    elif event.signal == signals.RIGHT_CLICK:
        status = chart.trans(in_project)
        chart.save_current_polygon_as_is()
    else:
        status = return_status.SUPER
        chart.temp.fun = in_project

    return status


@spy_on
def moving_point(chart: Statechart, event: Event) -> return_status:
    status = return_status.UNHANDLED

    if event.signal == signals.ENTRY_SIGNAL:
        status = return_status.HANDLED
        chart.on_moving_point_entry()
    if event.signal == signals.EXIT_SIGNAL:
        status = return_status.HANDLED
        chart.on_moving_point_exit()
    if event.signal == signals.UPDATE_FIGURE_POINT_POSITION:
        status = return_status.HANDLED
        chart.on_moving_point_update_figure_point_position(event.payload['figure_idx'],
                                                           event.payload['point_idx'],
                                                           event.payload['coords'])
    else:
        status = return_status.SUPER
        chart.temp.fun = in_project

    return status


@spy_on
def removing_point(chart: Statechart, event: Event) -> return_status:
    status = return_status.UNHANDLED

    if event.signal == signals.ENTRY_SIGNAL:
        status = return_status.HANDLED
        chart.on_removing_point_entry()
    if event.signal == signals.EXIT_SIGNAL:
        status = return_status.HANDLED
        chart.on_removing_point_exit()
    if event.signal == signals.UPDATE_FIGURE_REMOVE_POINT:
        status = return_status.HANDLED
        chart.on_removing_point_update_figure_remove_point(event.payload['figure_idx'], event.payload['point_idx'])
    else:
        status = return_status.SUPER
        chart.temp.fun = in_project

    return status


@spy_on
def adding_point(chart: Statechart, event: Event) -> return_status:
    status = return_status.UNHANDLED

    if event.signal == signals.ENTRY_SIGNAL:
        status = return_status.HANDLED
        chart.on_adding_point_entry()
    if event.signal == signals.EXIT_SIGNAL:
        status = return_status.HANDLED
        chart.on_adding_point_exit()
    if event.signal == signals.UPDATE_FIGURE_INSERT_POINT:
        status = return_status.HANDLED
        chart.on_adding_point_update_figure_insert_point(event.payload['figure_idx'], event.payload['point_idx'], event.payload['coords'])
    else:
        status = return_status.SUPER
        chart.temp.fun = in_project

    return status



import copy
import typing


class History:
    def __init__(self, defaults: typing.Dict = None):
        self.history = {}

        if defaults:
            for file_id, new_state in defaults.items():
                self.set_defaults(file_id, new_state)

    def set_defaults(self, file_id, new_state):
        self.history[file_id] = [copy.deepcopy(new_state)]

    def add_snapshot(self, file_id, new_state):
        self.history[file_id].append(copy.deepcopy(new_state))

    def pop_history(self, file_id):
        if len(self.history[file_id]) > 0:
            return self.history[file_id].pop()

    def has_history(self, file_id):
        if self.get_history_len(file_id) > 0:
            return True
        return False

    def get_history_len(self, file_id):
        if file_id in self.history:
            return len(self.history[file_id])
        return 0

    def reset_history(self):
        for file_id, file_states in self.history.items():
            self.set_defaults(file_id, file_states[-1])



import tkinter
from tkinter import ttk
import unittest

from image_label.gui import Gui
from image_label.common_bus import CommonBus


class TestGUIStructure(unittest.TestCase):
    def setUp(self):
        self.bus = CommonBus()
        self.gui = Gui(bus=self.bus)

    def test_gui_has_main_window(self):
        self.assertIsNotNone(self.gui.root)

    def test_gui_has_drawing_frame(self):
        self.assertIsNotNone(self.gui.drawing_frame)

    def test_drawing_frame_has_canvas(self):
        drawing_frame_has_canvas_child = False
        for child in self.gui.drawing_frame.winfo_children():
            if isinstance(child, tkinter.Canvas):
                drawing_frame_has_canvas_child = True
                break

        assert drawing_frame_has_canvas_child, 'Drawing frame has no canvas child'

    def test_gui_has_files_frame(self):
        self.assertIsNotNone(self.gui.files_frame, 'GUI has no Files frame')

    def test_files_frame_has_treeview(self):
        files_frame_has_treeview = False
        for child in self.gui.files_frame.winfo_children():
            if isinstance(child, ttk.Treeview):
                files_frame_has_treeview = True

        assert files_frame_has_treeview, 'Files frame has no Treeview'

    def test_files_treeview_contains_items_after_refresh(self):
        files = {
            '1': {'abs_path': '1', "filename": '1', "figures": []},
            '2': {'abs_path': '2', "filename": '2', "figures": []},
            '3': {'abs_path': '3', "filename": '3', "figures": []}
        }

        for file_id, file_data in files.items():
            self.gui.add_file(file_id, file_data)

        assert 3 == len(self.gui.files_frame_treeview.get_children())

        self.gui.clear_files()

        assert 0 == len(self.gui.files_frame_treeview.get_children())

        files = {
            '1': {'abs_path': '1', "filename": '1', "figures": []},
            '3': {'abs_path': '3', "filename": '3', "figures": []}
        }

        for file_id, file_data in files.items():
            self.gui.add_file(file_id, file_data)

        assert 2 == len(self.gui.files_frame_treeview.get_children())

    def test_gui_has_figures_frame(self):
        self.assertIsNotNone(self.gui.figures_frame, 'GUI has not Figures frame')

    def test_figures_frame_has_treeview(self):
        figures_frame_has_treeview = False
        for child in self.gui.figures_frame.winfo_children():
            if isinstance(child, ttk.Treeview):
                figures_frame_has_treeview = True

        assert figures_frame_has_treeview, 'Figures frame has no Treeview'


if __name__ == '__main__':
    unittest.main()



import tempfile
import unittest
from image_label.history import History


class TestHistory(unittest.TestCase):
    def test_create_history_with_no_defaults(self):
        h = History()

        assert isinstance(h, History)

    def test_create_history_with_some_defaults(self):
        data = {"1": {"abs_path": None}}

        h = History(defaults=data)

        assert isinstance(h, History)
        assert h.history != dict()
        assert len(h.history) == len(data)

    def test_create_empty_history_and_add_some_defaults(self):
        data = {"1": {"abs_path": None}}

        h = History()

        assert isinstance(h, History)
        assert h.history == dict()
        assert len(h.history) == 0

        for k, v in data.items():
            h.set_defaults(k, v)

        assert h.history != dict()
        assert len(h.history) == len(data)

    def test_create_empty_history_and_add_snapshot(self):
        data = {"1": {"abs_path": tempfile.mktemp()}}
        data2 = {"1": {"abs_path": tempfile.mktemp(), "figures": []}}

        h = History(data)

        for k, v in data2.items():
            h.add_snapshot(k, v)

        assert h.history != dict()
        assert h.pop_history('1') == data2['1']

    def test_create_empty_history_and_check_for_snapshots(self):
        h = History()

        assert not h.has_history('1')

    def test_create_not_empty_history_and_check_for_snapshots(self):
        data = {"1": {"abs_path": tempfile.mktemp()}}

        h = History(data)

        assert h.has_history('1')
        assert not h.has_history('2')

    def test_create_not_empty_history_and_check_pop_history(self):
        data = {"1": {"abs_path": tempfile.mktemp()}}

        h = History(data)

        assert h.pop_history('1') == data['1']


if __name__ == '__main__':
    unittest.main()



import re
import typing


class CommonBus:
    def __init__(self):
        self._items = {}

    @staticmethod
    def _check_if_array_and_return_groups(name: str):
        regex = r'^([a-zA-Z0-9\-_]*)\[([a-zA-Z0-9\-_]*)\]$'
        matches = re.findall(regex, name, re.MULTILINE | re.IGNORECASE | re.UNICODE)

        for match in matches:
            if len(match) == 2:
                return match

    def register_item(self, name: str, item: typing.Any):

        arr_data = self._check_if_array_and_return_groups(name)
        if arr_data:
            if arr_data[0] not in self._items:
                self._items[arr_data[0]] = {}
            self._items[arr_data[0]][arr_data[1]] = item
        else:
            self._items[name] = item

    def __getitem__(self, key):
        if key in self._items:
            return self._items[key]

    def __getattr__(self, key):
        if key in self._items:
            return self._items[key]



from image_label.common_bus import CommonBus
from image_label.gui import Gui
from image_label.statechart import Statechart
from image_label.exporters.yolo import YoloExporter


def run():
    b = CommonBus()
    s = Statechart('statechart', bus=b)
    y = YoloExporter(bus=b)
    g = Gui(bus=b)

    s.run()
    g.run()


if __name__ == '__main__':
    run()



import pathlib
import time
import unittest

from miros import stripped

from image_label import helpers

from image_label.common_bus import CommonBus
from image_label.statechart import Statechart
from image_label.gui import PlaceholderGui


class TestStates(unittest.TestCase):
    @staticmethod
    def _assert_trace_check(actual_trace, expected_trace):
        with stripped(expected_trace) as _expected_trace, stripped(actual_trace) as _actual_trace:
            assert len(_expected_trace) == len(
                _actual_trace), f'Not enough events: expected ({len(_expected_trace)}) != actual({len(_actual_trace)})'
            for expected, actual in zip(_expected_trace, _actual_trace):
                assert expected == actual, f'{expected} != {actual}'

    @staticmethod
    def _assert_spy_check(actual_spy, expected_spy):
        assert len(expected_spy) == len(actual_spy), f'Not enough events: expected ({len(expected_spy)}) != actual({len(actual_spy)})'
        for expected, actual in zip(actual_spy, expected_spy):
            assert expected == actual, f'{expected} != {actual}'

    def setUp(self):
        self.b = CommonBus()
        self.s = Statechart('statechart', bus=self.b)
        self.g = PlaceholderGui(bus=self.b)

        # self.s.live_spy = True
        # self.s.live_trace = True

        self.s.run()
        self.g.run()

    def test_from_no_project_to_in_project_on_new_project(self):
        helpers.new_project_event(self.s)
        time.sleep(0.1)

        expected_states = '''
        [2024-06-28 14:09:37.319643] [statechart] e->start_at() top->no_project
        [2024-06-28 14:09:37.319643] [statechart] e->NEW_PROJECT() no_project->in_project
        '''
        actual_states = self.s.trace()

        self._assert_trace_check(actual_states, expected_states)

    def test_from_no_project_to_in_project_on_load_project(self):
        helpers.load_project_event(self.s, pathlib.Path('./assets/domik.boobalp'))
        time.sleep(0.1)

        expected_states = '''
        [2024-06-28 15:14:35.233332] [statechart] e->start_at() top->no_project
        [2024-06-28 15:14:41.406999] [statechart] e->LOAD_PROJECT() no_project->in_project
        '''
        actual_states = self.s.trace()

        self._assert_trace_check(actual_states, expected_states)

    def test_from_no_project_to_in_project_on_new_project_on_add_file(self):
        helpers.new_project_event(self.s)
        time.sleep(0.1)
        helpers.add_file_event(self.s, [pathlib.Path('./assets/domiki.png').absolute()])
        time.sleep(0.1)

        expected_spy = ['START', 'SEARCH_FOR_SUPER_SIGNAL:no_project', 'ENTRY_SIGNAL:no_project', 'INIT_SIGNAL:no_project', '<- Queued:(0) Deferred:(0)', 'NEW_PROJECT:no_project', 'SEARCH_FOR_SUPER_SIGNAL:in_project', 'ENTRY_SIGNAL:in_project', 'INIT_SIGNAL:in_project', '<- Queued:(0) Deferred:(0)', 'ADD_FILE:in_project', 'POST_FIFO:SELECT_IMAGE', 'ADD_FILE:in_project:HOOK', '<- Queued:(1) Deferred:(0)', 'SELECT_IMAGE:in_project', 'POST_FIFO:RESET_DRAWING', 'SELECT_IMAGE:in_project:HOOK', '<- Queued:(1) Deferred:(0)', 'RESET_DRAWING:in_project', 'RESET_DRAWING:no_project', '<- Queued:(0) Deferred:(0)']
        actual_spy = self.s.spy()
        self._assert_spy_check(actual_spy, expected_spy)

    def test_project_no_project_to_in_project_on_new_project_on_add_file_on_remove_file(self):
        helpers.new_project_event(self.s)
        time.sleep(0.1)
        helpers.add_file_event(self.s, [pathlib.Path('./assets/domiki.png').absolute()])
        time.sleep(0.1)

        selected_file_id, _ = self.s.project.get_selected_file()
        helpers.remove_file_event(self.s, file_id=selected_file_id, force=True)
        time.sleep(0.1)

        expected_spy = ['START', 'SEARCH_FOR_SUPER_SIGNAL:no_project', 'ENTRY_SIGNAL:no_project', 'INIT_SIGNAL:no_project', '<- Queued:(0) Deferred:(0)', 'NEW_PROJECT:no_project', 'SEARCH_FOR_SUPER_SIGNAL:in_project', 'ENTRY_SIGNAL:in_project', 'INIT_SIGNAL:in_project', '<- Queued:(0) Deferred:(0)', 'ADD_FILE:in_project', 'POST_FIFO:SELECT_IMAGE', 'ADD_FILE:in_project:HOOK', '<- Queued:(1) Deferred:(0)', 'SELECT_IMAGE:in_project', 'POST_FIFO:RESET_DRAWING', 'SELECT_IMAGE:in_project:HOOK', '<- Queued:(1) Deferred:(0)', 'RESET_DRAWING:in_project', 'RESET_DRAWING:no_project', '<- Queued:(0) Deferred:(0)', 'REMOVE_FILE:in_project', 'REMOVE_FILE:in_project:HOOK', '<- Queued:(0) Deferred:(0)']
        actual_spy = self.s.spy()

        self._assert_spy_check(actual_spy, expected_spy)

    def test_project_no_project_to_in_project_to_drawing_rect_on_new_project_on_add_file_on_draw_rect(self):
        helpers.new_project_event(self.s)
        time.sleep(0.1)
        helpers.add_file_event(self.s, [pathlib.Path('./assets/domiki.png').absolute()])
        time.sleep(0.1)
        helpers.draw_rect_event(self.s)
        time.sleep(0.1)

        expected_states = '''
        [2024-06-28 15:55:31.905542] [statechart] e->start_at() top->no_project
        [2024-06-28 15:55:34.006727] [statechart] e->NEW_PROJECT() no_project->in_project
        [2024-06-28 15:55:38.815240] [statechart] e->DRAW_RECT() in_project->drawing_rect
        '''
        actual_states = self.s.trace()

        self._assert_trace_check(actual_states, expected_states)

    def test_project_no_project_to_in_project_to_drawing_rect_to_drawing_rect_waiting_for_2_point_to_in_project_on_new_project_on_add_file_on_draw_rect_on_click_on_click(self):
        helpers.new_project_event(self.s)
        time.sleep(0.1)
        helpers.add_file_event(self.s, [pathlib.Path('./assets/domiki.png').absolute()])
        time.sleep(0.1)
        helpers.draw_rect_event(self.s)
        time.sleep(0.1)
        helpers.click_canvas_event(self.s, coords=(0, 0))
        time.sleep(0.1)
        helpers.click_canvas_event(self.s, coords=(0, 0))
        time.sleep(0.1)

        expected_states = '''
        [2024-06-28 15:58:43.797013] [statechart] e->start_at() top->no_project
        [2024-06-28 15:58:48.040171] [statechart] e->NEW_PROJECT() no_project->in_project
        [2024-06-28 15:58:56.777206] [statechart] e->DRAW_RECT() in_project->drawing_rect
        [2024-06-28 15:58:57.922233] [statechart] e->CLICK() drawing_rect->drawing_rect_waiting_for_2_point
        [2024-06-28 15:59:00.724389] [statechart] e->CLICK() drawing_rect_waiting_for_2_point->in_project
        '''
        actual_states = self.s.trace()

        self._assert_trace_check(actual_states, expected_states)

    def test_project_no_project_to_in_project_to_drawing_poly_on_new_project_on_add_file_on_draw_poly(self):
        helpers.new_project_event(self.s)
        time.sleep(0.1)
        helpers.add_file_event(self.s, [pathlib.Path('./assets/domiki.png').absolute()])
        time.sleep(0.1)
        helpers.draw_poly_event(self.s)
        time.sleep(0.1)

        expected_states = '''
        [2024-06-28 15:55:31.905542] [statechart] e->start_at() top->no_project
        [2024-06-28 15:55:34.006727] [statechart] e->NEW_PROJECT() no_project->in_project
        [2024-06-28 15:55:38.815240] [statechart] e->DRAW_POLY() in_project->drawing_poly
        '''
        actual_states = self.s.trace()

        self._assert_trace_check(actual_states, expected_states)

    def test_project_no_project_to_in_project_to_drawing_poly_to_in_project_on_new_project_on_add_file_on_draw_poly_on_click_on_click_on_click(self):
        helpers.new_project_event(self.s)
        time.sleep(0.1)
        helpers.add_file_event(self.s, [pathlib.Path('./assets/domiki.png').absolute()])
        time.sleep(0.1)
        helpers.draw_poly_event(self.s)
        time.sleep(0.1)
        helpers.click_canvas_event(self.s, coords=(0, 0))
        time.sleep(0.1)
        helpers.click_canvas_event(self.s, coords=(100, 0))
        time.sleep(0.1)
        helpers.click_canvas_event(self.s, coords=(100, 50))
        time.sleep(0.1)
        helpers.click_canvas_event(self.s, coords=(0, 0))
        time.sleep(0.1)

        expected_states = '''
        [2024-06-28 16:22:39.068617] [statechart] e->start_at() top->no_project
        [2024-06-28 16:22:40.689900] [statechart] e->NEW_PROJECT() no_project->in_project
        [2024-06-28 16:22:45.130358] [statechart] e->DRAW_POLY() in_project->drawing_poly
        [2024-06-28 16:22:48.828693] [statechart] e->CLICK() drawing_poly->in_project
        '''
        actual_states = self.s.trace()

        self._assert_trace_check(actual_states, expected_states)

    def test_no_project_to_in_project_on_load_project_on_select_image_on_remove_figure(self):
        helpers.load_project_event(self.s, pathlib.Path('./assets/domik.boobalp'))
        time.sleep(0.1)
        helpers.figure_selected_event(self.s, 1)
        time.sleep(0.1)
        helpers.delete_figure_event(self.s, 1)
        time.sleep(0.1)

        expected_spy = ['START', 'SEARCH_FOR_SUPER_SIGNAL:no_project', 'ENTRY_SIGNAL:no_project', 'INIT_SIGNAL:no_project', '<- Queued:(0) Deferred:(0)', 'LOAD_PROJECT:no_project', 'SEARCH_FOR_SUPER_SIGNAL:in_project', 'ENTRY_SIGNAL:in_project', 'POST_FIFO:SELECT_IMAGE', 'INIT_SIGNAL:in_project', '<- Queued:(1) Deferred:(0)', 'SELECT_IMAGE:in_project', 'POST_FIFO:RESET_DRAWING', 'SELECT_IMAGE:in_project:HOOK', '<- Queued:(1) Deferred:(0)', 'RESET_DRAWING:in_project', 'RESET_DRAWING:no_project', '<- Queued:(0) Deferred:(0)', 'FIGURE_SELECTED:in_project', 'FIGURE_SELECTED:in_project:HOOK', '<- Queued:(0) Deferred:(0)', 'DELETE_FIGURE:in_project', 'DELETE_FIGURE:in_project:HOOK', '<- Queued:(0) Deferred:(0)']
        actual_spy = self.s.spy()

        self._assert_spy_check(actual_spy, expected_spy)

    def test_no_project_to_in_project_on_load_project_on_select_image_on_rotate_cw(self):
        helpers.load_project_event(self.s, pathlib.Path('./assets/domik.boobalp'))
        time.sleep(0.1)
        helpers.rotate_cw_event(self.s)
        time.sleep(0.1)

        expected_spy = ['START', 'SEARCH_FOR_SUPER_SIGNAL:no_project', 'ENTRY_SIGNAL:no_project', 'INIT_SIGNAL:no_project', '<- Queued:(0) Deferred:(0)', 'LOAD_PROJECT:no_project', 'SEARCH_FOR_SUPER_SIGNAL:in_project', 'ENTRY_SIGNAL:in_project', 'POST_FIFO:SELECT_IMAGE', 'INIT_SIGNAL:in_project', '<- Queued:(1) Deferred:(0)', 'SELECT_IMAGE:in_project', 'POST_FIFO:RESET_DRAWING', 'SELECT_IMAGE:in_project:HOOK', '<- Queued:(1) Deferred:(0)', 'RESET_DRAWING:in_project', 'RESET_DRAWING:no_project', '<- Queued:(0) Deferred:(0)', 'ROTATE_CW:in_project', 'ROTATE_CW:in_project:HOOK', '<- Queued:(0) Deferred:(0)']

        actual_spy = self.s.spy()

        self._assert_spy_check(actual_spy, expected_spy)

    def test_no_project_to_in_project_on_load_project_on_select_image_on_rotate_ccw(self):
        helpers.load_project_event(self.s, pathlib.Path('./assets/domik.boobalp'))
        time.sleep(0.1)
        helpers.rotate_ccw_event(self.s)
        time.sleep(0.1)

        expected_spy = ['START', 'SEARCH_FOR_SUPER_SIGNAL:no_project', 'ENTRY_SIGNAL:no_project', 'INIT_SIGNAL:no_project', '<- Queued:(0) Deferred:(0)', 'LOAD_PROJECT:no_project', 'SEARCH_FOR_SUPER_SIGNAL:in_project', 'ENTRY_SIGNAL:in_project', 'POST_FIFO:SELECT_IMAGE', 'INIT_SIGNAL:in_project', '<- Queued:(1) Deferred:(0)', 'SELECT_IMAGE:in_project', 'POST_FIFO:RESET_DRAWING', 'SELECT_IMAGE:in_project:HOOK', '<- Queued:(1) Deferred:(0)', 'RESET_DRAWING:in_project', 'RESET_DRAWING:no_project', '<- Queued:(0) Deferred:(0)', 'ROTATE_CCW:in_project', 'ROTATE_CCW:in_project:HOOK', '<- Queued:(0) Deferred:(0)']

        actual_spy = self.s.spy()

        self._assert_spy_check(actual_spy, expected_spy)

    def test_no_project_to_in_project_on_load_project_on_update_quick_categories(self):
        helpers.load_project_event(self.s, pathlib.Path('./assets/domik.boobalp'))
        time.sleep(0.1)

        assert self.s.project.get_quick_categories() is None

        new_categories = ['1', '2', '3']
        helpers.update_quick_categories(self.s, new_categories)
        time.sleep(0.1)

        assert self.s.project.get_quick_categories() == new_categories

        expected_spy = ['START', 'SEARCH_FOR_SUPER_SIGNAL:no_project', 'ENTRY_SIGNAL:no_project', 'INIT_SIGNAL:no_project', '<- Queued:(0) Deferred:(0)', 'LOAD_PROJECT:no_project', 'SEARCH_FOR_SUPER_SIGNAL:in_project', 'ENTRY_SIGNAL:in_project', 'POST_FIFO:SELECT_IMAGE', 'INIT_SIGNAL:in_project', '<- Queued:(1) Deferred:(0)', 'SELECT_IMAGE:in_project', 'POST_FIFO:RESET_DRAWING', 'SELECT_IMAGE:in_project:HOOK', '<- Queued:(1) Deferred:(0)', 'RESET_DRAWING:in_project', 'RESET_DRAWING:no_project', '<- Queued:(0) Deferred:(0)', 'UPDATE_QUICK_CATEGORIES:in_project', 'UPDATE_QUICK_CATEGORIES:in_project:HOOK', '<- Queued:(0) Deferred:(0)']

        actual_spy = self.s.spy()

        self._assert_spy_check(actual_spy, expected_spy)



if __name__ == '__main__':
    unittest.main()


import json
import pathlib
import tempfile
import unittest

from image_label.project import Project


class TestProject(unittest.TestCase):
    def test_create_empty_project(self):
        p = Project()

        assert isinstance(p, Project)

    def test_after_create_empty_project_object_has_files_dictionary(self):
        p = Project()

        assert isinstance(p.files, dict)

    def test_load_project(self):
        path = pathlib.Path('.', 'assets', 'empty.boobalp')
        p = Project(path)

        assert isinstance(p, Project)

    def test_after_load_empty_project_object_has_files_dictionary_and_contains_some_data(self):
        path = pathlib.Path('.', 'assets', 'empty.boobalp')

        p = Project(path)

        assert p.files == dict()

    def test_after_load_not_empty_project_object_has_files_dictionary_and_contains_some_data(self):
        path = pathlib.Path('.', 'assets', 'domik.boobalp')

        p = Project(path)

        assert p.files != dict()

        with open(path, 'r') as f:
            assert p.files == json.loads(f.read())['files']

    def test_after_load_not_empty_project_object_has_not_empty_history(self):
        path = pathlib.Path('.', 'assets', 'domik.boobalp')

        p = Project(path)

        assert p.history.has_history('img1')

    def test_after_adding_file_in_also_add_to_history(self):
        abs_path = tempfile.mktemp()
        p = Project()
        p.add_file(abs_path)

        assert p.history.has_history(list(p.files.keys())[0])

    def test_get_files_after_creating_not_empty_project(self):
        path = pathlib.Path('.', 'assets', 'domik.boobalp')

        p = Project(path)

        assert len(p.get_files()) > 0

    def test_select_file(self):
        abs_path = tempfile.mktemp()
        p = Project()
        p.add_file(abs_path)

        p.select_file(list(p.files.keys())[0])
        selected_file_id, selected_file_data = p.get_selected_file()
        assert selected_file_id == list(p.files.keys())[0]

    def test_delete_figure(self):
        path = pathlib.Path('.', 'assets', 'domik.boobalp')

        p = Project(path)
        p.select_file('img1')

        _, selected_file_data = p.get_selected_file()
        prev_number_of_figures = len(selected_file_data['figures'])

        p.delete_figure(1)
        _, selected_file_data = p.get_selected_file()

        assert len(selected_file_data['figures']) != prev_number_of_figures

    def test_save_object(self):
        file_path = tempfile.mktemp()
        p = Project()
        file_id = p.add_file(file_path)
        assert p.history.has_history(file_id)

        project_path = tempfile.mktemp()
        p.save_project(project_path)

        assert pathlib.Path(project_path).exists()

    def test_history_sets_to_1_after_saveing(self):
        path = pathlib.Path('.', 'assets', 'domik.boobalp')

        p = Project(path)
        p.select_file('img1')

        selected_file_id, selected_file_data = p.get_selected_file()

        for figure_id, _ in enumerate(selected_file_data['figures']):
            p.delete_figure(figure_id)

        assert p.history.get_history_len(selected_file_id) != 0
        assert p.history.get_history_len(selected_file_id) == 2

        project_path = tempfile.mktemp()
        p.save_project(project_path)

        assert p.history.get_history_len(selected_file_id) == 1

    def test_remove_file_from_project(self):
        path = pathlib.Path('.', 'assets', 'domik.boobalp')

        p = Project(path)
        number_of_files_before = len(p.get_files())
        expected_number_of_files_after = number_of_files_before - 1
        p.remove_file('img1')

        assert len(p.get_files()) == expected_number_of_files_after

    def test_get_files(self):
        path = pathlib.Path('.', 'assets', 'domik.boobalp')

        p = Project(path)
        keys = ('img2',)

        assert len(p.get_files()) != len(p.get_files(keys))
        assert len(p.get_files(keys)) == 1

    def test_update_figure_category(self):
        path = pathlib.Path('.', 'assets', 'domik.boobalp')

        p = Project(path)
        file_id = 'img2'
        figure_id = 0

        file_data = p.get_files((file_id,))
        p.update_figure_category(file_id, figure_id, 'test')

        assert file_data[0][1]['figures'][figure_id]['category'] == 'test'

    def test_rotate_cw(self):
        path = pathlib.Path('.', 'assets', 'domik.boobalp')
        p = Project(path)
        file_id = 'img2'

        assert p.history.get_history_len(file_id) == 1

        p.rotate_cw(file_id)

        assert p.history.get_history_len(file_id) == 2

        assert p.files[file_id]['transformations']
        assert len(p.files[file_id]['transformations']) == 1
        assert p.files[file_id]['transformations'] == ['rotate_cw']

    def test_rotate_ccw(self):
        path = pathlib.Path('.', 'assets', 'domik.boobalp')
        p = Project(path)
        file_id = 'img2'

        assert p.history.get_history_len(file_id) == 1

        p.rotate_ccw(file_id)

        assert p.history.get_history_len(file_id) == 2

        assert p.files[file_id]['transformations']
        assert len(p.files[file_id]['transformations']) == 1
        assert p.files[file_id]['transformations'] == ['rotate_ccw']

    def test_if_get_files_only_files_is_not_none_only_then_return_some_files(self):
        path = pathlib.Path('.', 'assets', 'domik.boobalp')
        p = Project(path)

        assert len(p.get_files()) == len(p.files)
        assert len(p.get_files(only_keys=None)) == len(p.files)
        assert len(p.get_files(only_keys=[])) == 0
        assert len(p.get_files(only_keys=['img2'])) == 1

    def test_project_set_project_quick_categories(self):
        p = Project()

        categories = ['a', 'b', 'c']
        p.set_quick_categories(categories)

        assert p.get_quick_categories() == categories


if __name__ == '__main__':
    unittest.main()
